"""Utility methods for working with the sensitivity task

"""
import logging

import basico
import COPASI

from . import model_io

logger = logging.getLogger(__name__)


class EnumHelper:
    """Utility class for dealing with mapping from names to enums easier

    """

    _names = None
    _values = None

    @classmethod
    def _create_name_map(cls):
        return dict()

    @classmethod
    def _create_value_map(cls):
        if cls._names is None:
            cls._names = cls._create_name_map()

        return basico.model_info._invert_dict(cls._names)

    @classmethod
    def _convert_names_to_values(cls):
        cls._values = basico.model_info._invert_dict(cls._names)

    @classmethod
    def from_enum(cls, int_value):
        if cls._names is None:
            cls._names = cls._create_name_map()

        return cls._names.get(int_value, next(iter(cls._names.items()))[1])

    @classmethod
    def to_enum(cls, value):
        if cls._values is None:
            cls._values = cls._create_value_map()

        return cls._values.get(value, next(iter(cls._values.items()))[1])

    @classmethod
    def get_all_names(cls):
        if cls._names is None:
            cls._names = cls._create_name_map()
        return cls._names.values()


class SENS(EnumHelper):
    """Class for Sensitivity object lists

    """

    EMPTY_LIST = "Not Set"
    SINGLE_OBJECT = "Single Object"

    METABS = "Species"
    METAB_INITIAL_CONCENTRATIONS = "Initial Concentrations"
    METAB_INITIAL_NUMBERS = "Initial Numbers"
    METAB_CONCENTRATIONS = "Concentrations of Species"
    METAB_NUMBERS = "Numbers of Species"
    NON_CONST_METAB_CONCENTRATIONS = "Non-Constant Concentrations of Species"
    NON_CONST_METAB_NUMBERS = "Non-Constant Numbers of Species"
    CONST_METAB_CONCENTRATIONS = "Constant Concentrations of Species"
    CONST_METAB_NUMBERS = "Constant Numbers of Species"
    ODE_METAB_CONCENTRATIONS = "Concentrations of Species with ODE"
    ODE_METAB_NUMBERS = "Numbers of Species with ODE"
    REACTION_METAB_CONCENTRATIONS = "Concentrations of Species determined by Reactions"
    REACTION_METAB_NUMBERS = "Numbers of Species determined by Reactions"
    ASS_METAB_CONCENTRATIONS = "Concentrations of Species with Assignment"
    ASS_METAB_NUMBERS = "Numbers of Species with Assignment"
    METAB_CONC_RATES = "Concentration Rates"
    METAB_PART_RATES = "Particle Rates"
    METAB_TRANSITION_TIME = "Transition Time"

    REACTIONS = "Reactions"
    REACTION_CONC_FLUXES = "Concentration Fluxes of Reactions"
    REACTION_PART_FLUXES = "Particle Fluxes of Reactions"

    GLOBAL_PARAMETERS = "Global Quantity"
    GLOBAL_PARAMETER_INITIAL_VALUES = "Global Quantity Initial Values"
    GLOBAL_PARAMETER_VALUES = "Global Quantity Values"
    NON_CONST_GLOBAL_PARAMETER_VALUES = "Non-Constant Global Quantity Values"
    CONST_GLOBAL_PARAMETER_INITIAL_VALUES = "Constant Global Quantity Values"
    ODE_GLOBAL_PARAMETER_VALUES = "Values of Global Quantities with ODE"
    ASS_GLOBAL_PARAMETER_VALUES = "Values of Global Quantities with Assignment"
    GLOBAL_PARAMETER_RATES = "Global Quantity Rates"

    COMPARTMENTS = "Compartments"
    COMPARTMENT_INITIAL_VOLUMES = "Compartment Initial Volumes"
    COMPARTMENT_VOLUMES = "Compartment Volumes"
    NON_CONST_COMPARTMENT_VOLUMES = "Non-Constant Compartment Volumes"
    CONST_COMPARTMENT_VOLUMES = "Constant Compartment Volumes"
    ODE_COMPARTMENT_VOLUMES = "Values of Compartment Volumes with ODE"
    ASS_COMPARTMENT_VOLUMES = "Values of Compartment Volumes with Assignment"
    COMPARTMENT_RATES = "Compartment Volume Rates"

    ALL_INITIAL_VALUES = "All initial Values"
    ALL_LOCAL_PARAMETER_VALUES = "Local Parameter Values"

    ALL_PARAMETER_VALUES = "All Parameter Values"
    ALL_PARAMETER_AND_INITIAL_VALUES = "All Parameter and Initial Values"

    ALL_VARIABLES = "All Variables of the model"
    ALL_ODE_VARIABLES = "All independent Variables of the model"

    REDUCED_JACOBIAN_EV_RE = "Real part of eigenvalues of the reduced jacobian"
    REDUCED_JACOBIAN_EV_IM = "Imaginary part of eigenvalues of the reduced jacobian"

    @classmethod
    def _create_name_map(cls):
        result = {
            COPASI.CObjectLists.EMPTY_LIST: SENS.EMPTY_LIST,
            COPASI.CObjectLists.SINGLE_OBJECT: SENS.SINGLE_OBJECT,

            COPASI.CObjectLists.METABS: SENS.METABS,
            COPASI.CObjectLists.METAB_INITIAL_CONCENTRATIONS: SENS.METAB_INITIAL_CONCENTRATIONS,
            COPASI.CObjectLists.METAB_INITIAL_NUMBERS: SENS.METAB_INITIAL_NUMBERS,
            COPASI.CObjectLists.METAB_CONCENTRATIONS: SENS.METAB_CONCENTRATIONS,
            COPASI.CObjectLists.METAB_NUMBERS: SENS.METAB_NUMBERS,
            COPASI.CObjectLists.NON_CONST_METAB_CONCENTRATIONS: SENS.NON_CONST_METAB_CONCENTRATIONS,
            COPASI.CObjectLists.NON_CONST_METAB_NUMBERS: SENS.NON_CONST_METAB_NUMBERS,
            COPASI.CObjectLists.CONST_METAB_CONCENTRATIONS: SENS.CONST_METAB_CONCENTRATIONS,
            COPASI.CObjectLists.CONST_METAB_NUMBERS: SENS.CONST_METAB_NUMBERS,
            COPASI.CObjectLists.ODE_METAB_CONCENTRATIONS: SENS.ODE_METAB_CONCENTRATIONS,
            COPASI.CObjectLists.ODE_METAB_NUMBERS: SENS.ODE_METAB_NUMBERS,
            COPASI.CObjectLists.REACTION_METAB_CONCENTRATIONS: SENS.REACTION_METAB_CONCENTRATIONS,
            COPASI.CObjectLists.REACTION_METAB_NUMBERS: SENS.REACTION_METAB_NUMBERS,
            COPASI.CObjectLists.ASS_METAB_CONCENTRATIONS: SENS.ASS_METAB_CONCENTRATIONS,
            COPASI.CObjectLists.ASS_METAB_NUMBERS: SENS.ASS_METAB_NUMBERS,
            COPASI.CObjectLists.METAB_CONC_RATES: SENS.METAB_CONC_RATES,
            COPASI.CObjectLists.METAB_PART_RATES: SENS.METAB_PART_RATES,
            COPASI.CObjectLists.METAB_TRANSITION_TIME: SENS.METAB_TRANSITION_TIME,

            COPASI.CObjectLists.REACTIONS: SENS.REACTIONS,
            COPASI.CObjectLists.REACTION_CONC_FLUXES: SENS.REACTION_CONC_FLUXES,
            COPASI.CObjectLists.REACTION_PART_FLUXES: SENS.REACTION_PART_FLUXES,

            COPASI.CObjectLists.GLOBAL_PARAMETERS: SENS.GLOBAL_PARAMETERS,
            COPASI.CObjectLists.GLOBAL_PARAMETER_INITIAL_VALUES: SENS.GLOBAL_PARAMETER_INITIAL_VALUES,
            COPASI.CObjectLists.GLOBAL_PARAMETER_VALUES: SENS.GLOBAL_PARAMETER_VALUES,
            COPASI.CObjectLists.NON_CONST_GLOBAL_PARAMETER_VALUES: SENS.NON_CONST_GLOBAL_PARAMETER_VALUES,
            COPASI.CObjectLists.CONST_GLOBAL_PARAMETER_INITIAL_VALUES: SENS.CONST_GLOBAL_PARAMETER_INITIAL_VALUES,
            COPASI.CObjectLists.ODE_GLOBAL_PARAMETER_VALUES: SENS.ODE_GLOBAL_PARAMETER_VALUES,
            COPASI.CObjectLists.ASS_GLOBAL_PARAMETER_VALUES: SENS.ASS_GLOBAL_PARAMETER_VALUES,
            COPASI.CObjectLists.GLOBAL_PARAMETER_RATES: SENS.GLOBAL_PARAMETER_RATES,

            COPASI.CObjectLists.COMPARTMENTS: SENS.COMPARTMENTS,
            COPASI.CObjectLists.COMPARTMENT_INITIAL_VOLUMES: SENS.COMPARTMENT_INITIAL_VOLUMES,
            COPASI.CObjectLists.COMPARTMENT_VOLUMES: SENS.COMPARTMENT_VOLUMES,
            COPASI.CObjectLists.NON_CONST_COMPARTMENT_VOLUMES: SENS.NON_CONST_COMPARTMENT_VOLUMES,
            COPASI.CObjectLists.CONST_COMPARTMENT_VOLUMES: SENS.CONST_COMPARTMENT_VOLUMES,
            COPASI.CObjectLists.ODE_COMPARTMENT_VOLUMES: SENS.ODE_COMPARTMENT_VOLUMES,
            COPASI.CObjectLists.ASS_COMPARTMENT_VOLUMES: SENS.ASS_COMPARTMENT_VOLUMES,
            COPASI.CObjectLists.COMPARTMENT_RATES: SENS.COMPARTMENT_RATES,

            COPASI.CObjectLists.ALL_INITIAL_VALUES: SENS.ALL_INITIAL_VALUES,
            COPASI.CObjectLists.ALL_LOCAL_PARAMETER_VALUES: SENS.ALL_LOCAL_PARAMETER_VALUES,

            COPASI.CObjectLists.ALL_PARAMETER_VALUES: SENS.ALL_PARAMETER_VALUES,
            COPASI.CObjectLists.ALL_PARAMETER_AND_INITIAL_VALUES: SENS.ALL_PARAMETER_AND_INITIAL_VALUES,

            COPASI.CObjectLists.ALL_VARIABLES: SENS.ALL_VARIABLES,
            COPASI.CObjectLists.ALL_ODE_VARIABLES: SENS.ALL_ODE_VARIABLES,

            COPASI.CObjectLists.REDUCED_JACOBIAN_EV_RE: SENS.REDUCED_JACOBIAN_EV_RE,
            COPASI.CObjectLists.REDUCED_JACOBIAN_EV_IM: SENS.REDUCED_JACOBIAN_EV_IM
        }

        cls._names = result
        cls._values = basico.model_info._invert_dict(result)

        return result


class SENS_ST(EnumHelper):
    """Enumeration of Sensitivity Subtasks"""
    Evaluation = "Evaluation"
    SteadyState = "Steady State"
    TimeSeries = "Time Series"
    ParameterEstimation = "Parameter Estimation"
    Optimization = "Optimization"
    CrossSection = "Cross Section"

    @classmethod
    def _create_name_map(cls):
        return {
            COPASI.CSensProblem.Evaluation: SENS_ST.Evaluation,
            COPASI.CSensProblem.SteadyState: SENS_ST.SteadyState,
            COPASI.CSensProblem.TimeSeries: SENS_ST.TimeSeries,
            COPASI.CSensProblem.ParameterEstimation: SENS_ST.ParameterEstimation,
            COPASI.CSensProblem.Optimization: SENS_ST.Optimization,
            COPASI.CSensProblem.CrossSection: SENS_ST.CrossSection,
        }

    @classmethod
    def _create_value_map(cls):
        cls._names = cls._create_name_map()
        cls._convert_names_to_values()

        # add alternative names used throughout basico elsewhere
        cls._values[basico.T.STEADY_STATE] = COPASI.CSensProblem.SteadyState
        cls._values[basico.T.TIME_COURSE] = COPASI.CSensProblem.TimeSeries

        return cls._values


def get_sensitivity_settings(basic_only=True, **kwargs):
    """Returns the settings of the sensitivity task

    :param basic_only: if true, only basic parameters will be returned
    :type basic_only: bool

    :param kwargs: optional arguments

         - | `model`: to specify the data model to be used (if not specified
           | the one from :func:`.get_current_model` will be taken)


    :return: the settings as dictionary
    :rtype: dict
    """
    model = model_io.get_model_from_dict_or_default(kwargs)
    assert (isinstance(model, COPASI.CDataModel))

    settings = basico.get_task_settings(basico.T.SENSITIVITIES, basic_only, model=model)
    del settings['problem']

    task = model.getTask(basico.T.SENSITIVITIES)
    assert (isinstance(task, COPASI.CSensTask))
    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CSensProblem))

    settings['sub_task'] = SENS_ST.from_enum(problem.getSubTaskType())

    item = problem.getTargetFunctions()
    assert (isinstance(item, COPASI.CSensItem))

    settings['effect'] = _get_name_for_item(item, model)

    num_causes = problem.getNumberOfVariables()

    cause = SENS.EMPTY_LIST
    if num_causes > 0:
        cause = _get_name_for_item(problem.getVariables(0), model)
    settings['cause'] = cause

    cause = SENS.EMPTY_LIST
    if num_causes > 1:
        cause = _get_name_for_item(problem.getVariables(1), model)
    settings['secondary_cause'] = cause

    return settings


def set_sensitivity_settings(settings, **kwargs):
    """Applies the settings to the sensitivity task

    :param settings: the setting dictionary with keys, as obtained by get_sensitivity_settings
    :type settings: dict

    :param kwargs: optional arguments

         - | `model`: to specify the data model to be used (if not specified
           | the one from :func:`.get_current_model` will be taken)


    :return: None
    """
    model = model_io.get_model_from_dict_or_default(kwargs)
    assert (isinstance(model, COPASI.CDataModel))

    basico.set_task_settings(basico.T.SENSITIVITIES, settings, model=model)

    task = model.getTask(basico.T.SENSITIVITIES)
    assert (isinstance(task, COPASI.CSensTask))
    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CSensProblem))

    if 'sub_task' in settings:
        problem.setSubTaskType(SENS_ST.to_enum(settings['sub_task']))

    if 'effect' in settings:
        problem.setTargetFunctions(_to_sens_item(settings['effect'], model))

    if 'cause' in settings:
        problem.changeVariables(0, _to_sens_item(settings['cause'], model))

    if 'secondary_cause' in settings:
        problem.changeVariables(1, _to_sens_item(settings['secondary_cause'], model))


def _to_sens_item(name, model):
    """converts the name to a sens item

    :param name: display name, cn or object list name
    :type name: str
    :param model: the model to resolve elements in
    :return: the sens item for the named element
    """
    if name in SENS.get_all_names():
        return COPASI.CSensItem(SENS.to_enum(name))

    if name.startswith('CN='):
        return COPASI.CSensItem(COPASI.CCommonName(name))

    obj = model.findObjectByDisplayName(name)
    if obj is None:
        logger.warning('Object could not be resolved {0}'.format(name))
        return COPASI.CSensItem(0)

    if obj.getObjectType() != 'Reference':
        obj = obj.getValueReference()

    return COPASI.CSensItem(obj.getCN())


def _get_name_for_item(item, model):
    """ Transforms the sens item to a name

    :param item: the sens item
    :type: COPASI.CSensItem

    :param model: the model for resolving elements in
    :type model: COPASI.CDataModel

    :return: the name for the element, or list
    :rtype: str
    """

    if item.isSingleObject():
        cn = item.getSingleObjectCN()
        obj = model.getObject(cn)
        if obj is None:
            logger.warning('Object for sensitivities not found')

        name = basico.model_info._get_name_for_object(obj, model=model)
    else:
        name = item.getListTypeDisplayName()
    return name


def run_sensitivities(**kwargs):
    """Runs the sensitivity task, the result is obtained

    by calling:

    * `get_scaled_sensitivies`
    * `get_unscaled_sensitivies` or
    * `get_summarized_sensitivites`

    :param kwargs: optional arguments

         - | `model`: to specify the data model to be used (if not specified
           | the one from :func:`.get_current_model` will be taken)

         - | `settings`: a dictionary with the settings to apply first

         - | `use_initial_values`: boolean flag indicating whether initial values should be used (true by default)

    :return: None
    """
    model = model_io.get_model_from_dict_or_default(kwargs)
    assert (isinstance(model, COPASI.CDataModel))

    model.getModel().compileIfNecessary()

    if 'settings' in kwargs:
        basico.set_sensitivity_settings(kwargs['settings'], model=model)

    task = model.getTask(basico.T.SENSITIVITIES)
    assert (isinstance(task, COPASI.CSensTask))
    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CSensProblem))

    if not task.initializeRaw(COPASI.CCopasiTask.OUTPUT_UI):
        logger.error('Could not initialize Sensitivity Task: {0}'.format(COPASI.CCopasiMessage.getAllMessageText()))
        return

    use_initial_values = kwargs.get('use_initial_values', True)

    if not task.processRaw(use_initial_values):
        logger.error('Could not run Sensitivity Task: {0}'.format(COPASI.CCopasiMessage.getAllMessageText()))
        return

    task.restore()


def get_scaled_sensitivities(**kwargs):
    """Returns the scaled sensitivity matrix as pandas data frame

    :param kwargs: optional arguments

         - | `model`: to specify the data model to be used (if not specified
           | the one from :func:`.get_current_model` will be taken)

         - | `run_first`: boolean flag indicating that the task should be run first (defaults to false)

         - | `settings`: a dictionary with the settings to apply first

         - | `use_initial_values`: boolean flag indicating whether initial values should be used (true by default)

    :return: the scaled sensitivity matrix
    :rtype: pd.DataFrame
    """

    model = model_io.get_model_from_dict_or_default(kwargs)
    assert (isinstance(model, COPASI.CDataModel))

    if 'run_first' in kwargs and kwargs['run_first']:
        run_sensitivities(**kwargs)

    task = model.getTask(basico.T.SENSITIVITIES)
    assert (isinstance(task, COPASI.CSensTask))
    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CSensProblem))

    return basico.model_info._annotated_matrix_to_df(problem.getScaledResultAnnotated())


def get_unscaled_sensitivities(**kwargs):
    """Returns the unscaled sensitivity matrix as pandas data frame

    :param kwargs: optional arguments

         - | `model`: to specify the data model to be used (if not specified
           | the one from :func:`.get_current_model` will be taken)

         - | `run_first`: boolean flag indicating that the task should be run first (defaults to false)

         - | `settings`: a dictionary with the settings to apply first

         - | `use_initial_values`: boolean flag indicating whether initial values should be used (true by default)

    :return: the unscaled sensitivity matrix
    :rtype: pd.DataFrame
    """

    model = model_io.get_model_from_dict_or_default(kwargs)
    assert (isinstance(model, COPASI.CDataModel))

    if 'run_first' in kwargs and kwargs['run_first']:
        run_sensitivities(**kwargs)

    task = model.getTask(basico.T.SENSITIVITIES)
    assert (isinstance(task, COPASI.CSensTask))
    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CSensProblem))

    return basico.model_info._annotated_matrix_to_df(problem.getResultAnnotated())


def get_summarized_sensitivities(**kwargs):
    """Returns the summarized sensitivity matrix as pandas data frame

    The summarized result is computed as the 2 norm of the scaled result, thus collapsing the dimension

    :param kwargs: optional arguments

         - | `model`: to specify the data model to be used (if not specified
           | the one from :func:`.get_current_model` will be taken)

         - | `run_first`: boolean flag indicating that the task should be run first (defaults to false)

         - | `settings`: a dictionary with the settings to apply first

         - | `use_initial_values`: boolean flag indicating whether initial values should be used (true by default)

    :return: the summarized sensitivity matrix
    :rtype: pd.DataFrame or None
    """

    model = model_io.get_model_from_dict_or_default(kwargs)
    assert (isinstance(model, COPASI.CDataModel))

    if 'run_first' in kwargs and kwargs['run_first']:
        run_sensitivities(**kwargs)

    task = model.getTask(basico.T.SENSITIVITIES)
    assert (isinstance(task, COPASI.CSensTask))
    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CSensProblem))

    return basico.model_info._annotated_matrix_to_df(problem.getCollapsedResultAnnotated())




""" Utility methods for working with the Lyapunov task
"""

import logging
from math import e
import basico
import COPASI

from . import model_io

logger = logging.getLogger(__name__)


def run_lyapunov(**kwargs):
    """Runs the lyapunov task, the result is obtained

    by calling:

    * `get_lyapunov_exponents`

    :param kwargs: optional arguments

         - | `model`: to specify the data model to be used (if not specified
           | the one from :func:`.get_current_model` will be taken)

         - | `num_exponents`: number of exponents to calculate (default: 3)

         - | `calculate_divergence`: boolean flag indicating whether the divergence should be calculated (default: True)

         - | `start_averaging_after`: time after which the averaging should start (default: disabled = 0)

         - | `settings`: a dictionary with the settings to apply first

         - | `use_initial_values`: boolean flag indicating whether initial values should be used (true by default)

    Example:

    .. code-block:: python

        >>> load_example('brusselator')
        >>> exponents, sums, divergence = run_lyapunov(num_exponents=2)
        >>> print(exponents)
        [-0.0007025645113957691, -2.6051942604518477]

    :return: tuple of exponents, sum of exponents and average divergence

    """
    model = model_io.get_model_from_dict_or_default(kwargs)
    assert (isinstance(model, COPASI.CDataModel))

    model.getModel().compileIfNecessary()

    if 'settings' in kwargs:
        basico.set_task_settings(basico.T.LYAPUNOV_EXPONENTS, kwargs['settings'], model=model)

    task = model.getTask(basico.T.LYAPUNOV_EXPONENTS)
    assert (isinstance(task, COPASI.CLyapTask))

    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CLyapProblem))

    if 'num_exponents' in kwargs:
        problem.setExponentNumber(int(kwargs['num_exponents']))

    if 'calculate_divergence' in kwargs:
        problem.setDivergenceRequested(bool(kwargs['calculate_divergence']))

    if 'start_averaging_after' in kwargs:
        problem.setTransientTime(float(kwargs['start_averaging_after']))

    if not task.initializeRaw(COPASI.CCopasiTask.OUTPUT_UI):
        logger.error('Could not initialize Lyapunov Task: {0}'.format(COPASI.CCopasiMessage.getAllMessageText()))
        return

    use_initial_values = kwargs.get('use_initial_values', True)

    if not task.processRaw(use_initial_values):
        logger.error('Could not run Lyapunov Task: {0}'.format(COPASI.CCopasiMessage.getAllMessageText()))
        return

    task.restore()

    return get_lyapunov_exponents(model=model)


def get_lyapunov_exponents(**kwargs):
    """Returns the lyapunov exponents calclated last time `run_lyapunov` was called

    :param kwargs: optional arguments

         - | `model`: to specify the data model to be used (if not specified
           | the one from :func:`.get_current_model` will be taken)

    :return: a tuple of exponent, sum of exponents and average divergence
    """
    model = model_io.get_model_from_dict_or_default(kwargs)
    
    assert (isinstance(model, COPASI.CDataModel))

    task = model.getTask(basico.T.LYAPUNOV_EXPONENTS)
    assert (isinstance(task, COPASI.CLyapTask))


    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CLyapProblem))

    num_exponents = task.numberOfExponentsCalculated()
    exponent_vec = task.exponents()
    exponents = []

    for i in range(num_exponents):
        exponents.append(exponent_vec.get(i))

    sum_exponents = task.sumOfExponents()
    
    average_divergence = task.averageDivergence()
    
    return exponents, sum_exponents, average_divergence


import unittest
import basico


class TestSensitivities(unittest.TestCase):
    def setUp(self):
        basico.load_example('bruss')
        self.assertIsNotNone(basico.get_species('X', exact=True))

    def test_settings(self):
        settings = basico.get_sensitivity_settings()
        self.assertDictEqual(settings,
                             {
                                 'scheduled': False,
                                 'update_model': False,
                                 'method': {'Delta factor': 0.001,
                                            'Delta minimum': 1e-12,
                                            'name': 'Sensitivities Method'},
                                 'report': {'filename': '',
                                            'report_definition': 'Sensitivities',
                                            'append': True,
                                            'confirm_overwrite': False},
                                 'sub_task': 'Steady State',
                                 'effect': 'Non-Constant Concentrations of Species',
                                 'cause': 'All Parameter Values',
                                 'secondary_cause': 'Not Set'
                              })

        basico.set_sensitivity_settings({
            'effect': basico.SENS.METAB_CONCENTRATIONS,
            'cause': basico.SENS.METAB_INITIAL_CONCENTRATIONS,
            'scheduled': True
        })

        settings = basico.get_sensitivity_settings()
        self.assertEqual(settings['scheduled'], True)
        self.assertEqual(settings['effect'], basico.SENS.METAB_CONCENTRATIONS)
        self.assertEqual(settings['cause'], basico.SENS.METAB_INITIAL_CONCENTRATIONS)

    def test_enums(self):
        self.assertEqual(basico.SENS.from_enum(1), basico.SENS.SINGLE_OBJECT)
        self.assertEqual(basico.SENS.to_enum(basico.SENS.SINGLE_OBJECT), 1)

        self.assertEqual(basico.SENS_ST.to_enum(basico.SENS_ST.Evaluation), 0)
        self.assertEqual(basico.SENS_ST.to_enum(basico.SENS_ST.SteadyState), 1)
        self.assertEqual(basico.SENS_ST.to_enum(basico.SENS_ST.TimeSeries), 2)
        self.assertEqual(basico.SENS_ST.to_enum(basico.T.TIME_COURSE), 2)
        self.assertEqual(basico.SENS_ST.to_enum(basico.T.TIME_COURSE),
                         basico.SENS_ST.to_enum(basico.SENS_ST.TimeSeries))

        self.assertEqual(basico.SENS.from_enum(3), basico.SENS.METAB_INITIAL_CONCENTRATIONS)

    def test_run(self):
        basico.run_sensitivities()
        scaled = basico.get_scaled_sensitivities()
        self.assertIsNotNone(scaled)
        unscaled = basico.get_unscaled_sensitivities()
        self.assertIsNotNone(unscaled)
        summarized = basico.get_summarized_sensitivities()
        self.assertIsNotNone(summarized)


if __name__ == '__main__':
    unittest.main()



import logging
from math import exp
import sys
import unittest
import basico
import COPASI


class TestBasicoIO_Brus(unittest.TestCase):

    def setUp(self):
        self.dm = basico.load_example('brusselator')
        self.assertTrue(self.dm is not None)
        self.assertTrue(isinstance(self.dm, COPASI.CDataModel))
        self.assertTrue('The Brusselator' in basico.model_io.overview())

    def test_get_species(self):
        species = basico.get_species('X')
        self.assertTrue(species.shape[0] == 1)

    def test_species_modification(self):
        basico.add_species('speciesB');
        basico.set_species('speciesB', exact=True, initial_concentration=0.4, status='fixed');
        species = basico.as_dict(basico.get_species('speciesB', exact=True))
        self.assertIsNotNone(species)
        self.assertAlmostEqual(species['initial_concentration'], 0.4)
        self.assertEqual(species['type'], 'fixed')

        basico.add_species('speciesC', initial_concentration=0.3, status='fixed');
        species = basico.as_dict(basico.get_species('speciesC', exact=True))
        self.assertIsNotNone(species)
        self.assertAlmostEqual(species['initial_concentration'], 0.3)
        self.assertEqual(species['type'], 'fixed')

        basico.add_species('speciesD', initial_particle_number=200, status='fixed');
        species = basico.as_dict(basico.get_species('speciesD', exact=True))
        self.assertIsNotNone(species)
        self.assertAlmostEqual(species['initial_particle_number'], 200)
        self.assertEqual(species['type'], 'fixed')

    def test_reaction_modifiers(self):
        basico.add_reaction('mod_r', 'species_A -> species_B', function='Allosteric inhibition (MWC)')
        reaction = basico.as_dict(basico.get_reactions('mod_r', exact=True))
        self.assertIsNotNone(reaction)
        # this should fail, since there is no modifier mapping
        self.assertNotEqual(reaction['function'], 'Allosteric inhibition (MWC)')

        # this should fail too, since the mapping does not apply
        basico.set_reaction('mod_r', function='Allosteric inhibition (MWC)', mapping={'non_existing': 'species_A'})
        reaction = basico.as_dict(basico.get_reactions('mod_r', exact=True))
        self.assertIsNotNone(reaction)
        self.assertNotEqual(reaction['function'], 'Allosteric inhibition (MWC)')

        # now again with inhibitor, this time it should work
        basico.set_reaction('mod_r', function='Allosteric inhibition (MWC)', mapping={'Inhibitor': 'species_A'})
        reaction = basico.as_dict(basico.get_reactions('mod_r', exact=True))
        self.assertIsNotNone(reaction)
        self.assertEqual(reaction['function'], 'Allosteric inhibition (MWC)')

    def test_get_reaction_parameters(self):
        parameters = basico.get_reaction_parameters('k1')
        self.assertTrue(parameters.shape[0] == 4)
        parameters = basico.get_reaction_parameters('(R1).k1')
        self.assertTrue(parameters.shape[0] == 1)

    def test_set_reaction_parameters(self):
        parameters = basico.get_reaction_parameters('(R1).k1')
        self.assertEqual(parameters.shape[0], 1)
        value = parameters.iloc[0].value
        self.assertEqual(value, 1)
        basico.set_reaction_parameters('(R1).k1', value=2)
        value = basico.get_reaction_parameters('(R1).k1').iloc[0].value
        self.assertEqual(value, 2.0)

    def test_timecourse(self):
        data = basico.run_time_course()
        self.assertTrue(data.shape[1] == 2)
        basico.set_model_unit(substance_unit=1)
        data_stoch = basico.run_time_course(method='stochastic')
        self.assertTrue(data_stoch.shape[1] == 2)

    def test_steadystate(self):
        result = basico.run_steadystate()
        self.assertTrue(result == 1)

    def test_expressions(self):
        result = basico.model_info._split_by_cn('<CN=Root,Model=model0,Vector=Compartments[compartment],'
                                                'Vector=Metabolites[P],Reference=Concentration>*<CN=Root,Model='
                                                'model0,Vector=Values[epsilon],Reference=Value>+<CN=Root,Model='
                                                'model0,Vector=Values[offset],Reference=Value>')
        self.assertTrue(len(result) == 5)
        self.assertTrue(result[0] == 'CN=Root,Model=model0,Vector=Compartments[compartment],Vector='
                                     'Metabolites[P],Reference=Concentration')
        self.assertTrue(result[1] == '*')

        result = basico.model_info._split_by_cn('sin(A + B + C)')
        self.assertTrue(len(result) == 8)

    def test_notes(self):
        notes = basico.model_info.get_notes()
        self.assertTrue('The famous Brusselator' in notes)
        basico.model_info.set_notes("""

        New Multiline Comment

        """)
        new_notes = basico.model_info.get_notes()
        self.assertTrue('Multiline' in new_notes)

        metab_notes = basico.model_info.get_notes(name='[X]')
        self.assertTrue(metab_notes.strip() == '')
        basico.model_info.set_notes('Notes on the species X',  name='[X]')
        new_metab_notes = basico.model_info.get_notes(name='[X]')
        self.assertTrue('species X' in new_metab_notes)

    def test_events(self):
        events = basico.model_info.get_events()
        self.assertIsNone(events)
        basico.add_event('e0', trigger='Time > 10', assignments=[('[X]', '10')])
        events = basico.get_events('e0').reset_index()
        self.assertIsNotNone(events)
        d = events.to_dict(orient='records')[0]
        self.assertEqual(d['name'], 'e0')
        self.assertEqual(d['trigger'], 'Time > 10')
        self.assertEqual(d['assignments'][0]['target'], '[X]')
        self.assertEqual(d['assignments'][0]['expression'], '10')
        basico.model_info.add_event_assignment('e0', assignment=('[Y]', '2'))
        events = basico.get_events('e0').reset_index()
        d = events.to_dict(orient='records')[0]
        self.assertEqual(len(d['assignments']), 2)

    def test_matrices(self):
        jac = basico.model_info.get_jacobian_matrix(True)
        self.assertIsNotNone(jac)
        red = basico.model_info.get_reduced_jacobian_matrix()
        self.assertIsNotNone(red)
        stoich = basico.model_info.get_stoichiometry_matrix()
        self.assertIsNotNone(stoich)
        red_s = basico.model_info.get_reduced_stoichiometry_matrix()
        self.assertIsNotNone(red_s)

    def test_new_getter(self):
        tasks = basico.model_info.get_scheduled_tasks()
        self.assertEqual(len(tasks), 0)
        basico.set_scheduled_tasks(basico.T.TIME_COURSE)
        tasks = basico.model_info.get_scheduled_tasks()
        self.assertEqual(len(tasks), 1)
        cn = basico.model_info.get_cn('[X]')
        self.assertTrue('Metabolite' in cn)
        val = basico.model_info.get_value('[X]')
        self.assertIsNotNone(val)

    def test_amount(self):
        params = basico.get_parameters(type='assignment')
        self.assertIsNone(params)
        basico.add_amount_expressions()
        params2 = basico.get_parameters(type='assignment')
        self.assertIsNotNone(params2)
        basico.remove_amount_expressions()
        param3 = basico.get_parameters(type='assignment')
        self.assertIsNone(param3)

    def test_get_plots(self):
        result = basico.model_info.get_plots()
        self.assertEqual(len(result), 3)

    def test_add_plot(self):
        basico.model_info.add_plot('test', curves=[{'color': '#FF0000', 'channels': ['Time', '[X]']}])
        result = basico.model_info.get_plots()
        self.assertEqual(len(result), 4)
        plot = basico.model_info.get_plot_dict('test')
        self.assertTrue(len(plot['curves']) == 1)
        self.assertTrue(len(plot['curves'][0]['channels']) == 2)
        self.assertEqual(plot['curves'][0]['color'], '#FF0000')
        basico.model_info.remove_plot('test')
        result = basico.model_info.get_plots()
        self.assertEqual(len(result), 3)

    def test_collect_data(self):
        basico.add_parameter('quantity', initial_value=1)
        mod = basico.get_current_model().getModel()
        assert (isinstance(mod, COPASI.CModel))
        mod.applyInitialValues()
        data = basico.model_info._collect_data(cns=[
            'CN=Root,Model=The Brusselator,Reference=Time',
            'CN=Root,Model=The Brusselator,Vector=Compartments[compartment],Reference=InitialVolume',
            'CN=Root,Model=The Brusselator,Vector=Compartments[compartment],Reference=Rate',
            'CN=Root,Model=The Brusselator,Vector=Compartments[compartment],Reference=Volume',
            'CN=Root,Model=The Brusselator,Vector=Compartments[compartment],Vector=Metabolites[X],Reference=InitialParticleNumber',
            'CN=Root,Model=The Brusselator,Vector=Compartments[compartment],Vector=Metabolites[X],Reference=ParticleNumber',
            'CN=Root,Model=The Brusselator,Vector=Compartments[compartment],Vector=Metabolites[X],Reference=ParticleNumberRate',
            'CN=Root,Model=The Brusselator,Vector=Compartments[compartment],Vector=Metabolites[X],Reference=Rate',
            'CN=Root,Model=The Brusselator,Vector=Compartments[compartment],Vector=Metabolites[X],Reference=Concentration',
            'CN=Root,Model=The Brusselator,Vector=Compartments[compartment],Vector=Metabolites[X],Reference=InitialConcentration',
            'CN=Root,Model=The Brusselator,Vector=Values[quantity],Reference=Value',
            'CN=Root,Model=The Brusselator,Vector=Values[quantity],Reference=InitialValue',
            'CN=Root,Model=The Brusselator,Vector=Values[quantity],Reference=Rate',
            'CN=Root,Model=The Brusselator,Vector=Reactions[R1],Reference=Flux',
            'CN=Root,Model=The Brusselator,Vector=Reactions[R1],Reference=ParticleFlux',
            'CN=Root,Model=The Brusselator,Vector=Reactions[R1],ParameterGroup=Parameters,Parameter=k1,Reference=Value'
        ])

        data2 = basico.model_info._collect_data(names=[
            'Time',
            'Compartments[compartment].InitialVolume',
            'Compartments[compartment].Rate',
            'Compartments[compartment].Volume',
            'X.InitialParticleNumber',
            'X.ParticleNumber',
            'X.InitialParticleNumberRate',
            'X.ParticleNumberRate',
            'X.Rate',
            '[X]',
            '[X]_0',
            'Values[quantity]',
            'Values[quantity].InitialValue',
            'Values[quantity].Rate',
            '(R1).Flux',
            '(R1).ParticleFlux',
            '(R1).k1'
        ])
        self.assertTrue(data2 is not None)


class TestBasicoIO_LM(unittest.TestCase):

    def setUp(self):
        self.dm = basico.load_example('LM-test1')
        self.assertTrue(self.dm is not None)
        self.assertTrue(isinstance(self.dm, COPASI.CDataModel))
        self.assertTrue('Kinetics of a  Michaelian enzyme measured spectrophotometrically' in basico.model_io.overview())

    def tearDown(self):
        basico.remove_datamodel(self.dm)

    def test_get_parameter(self):
        params = basico.get_parameters()
        self.assertEqual(params.shape[0], 3)
        params = basico.get_parameters('epsilon')
        self.assertEqual(params.shape[0], 1)
        value = params.iloc[0].initial_value
        self.assertEqual(value, 0.78)

    def test_set_parameter(self):
        params = basico.get_parameters('epsilon')
        self.assertEqual(params.shape[0], 1)
        value = params.iloc[0].initial_value
        self.assertEqual(value, 0.78)
        basico.set_parameters('epsilon', initial_value=2.0)
        params = basico.get_parameters('epsilon')
        self.assertEqual(params.shape[0], 1)
        value = params.iloc[0].initial_value
        self.assertEqual(value, 2.0)
        basico.set_parameters('Values[epsilon]', initial_value=3.0, value=5.0)
        params = basico.as_dict(basico.get_parameters('Values[epsilon]'))
        self.assertEqual(params['initial_value'], 3.0)
        self.assertEqual(params['value'], 5.0)

    def test_mapping(self):
        m = basico.get_reaction_mapping('R1')
        self.assertDictEqual(m, {'k1': 130.0, 'substrate': ['S', 'E'], 'k2': 1.0, 'product': 'ES'})
        basico.set_reaction_mapping('R1', {'k1': 10})
        m = basico.get_reaction_mapping('R1')
        self.assertEqual(m['k1'], 10)
        basico.add_reaction('R3', 'S + E + ES = C + D + F')
        m = basico.get_reaction_mapping('R3')
        self.assertListEqual(m['product'], ['C', 'D', 'F'])
        basico.set_reaction_mapping('R3', {'product': ['S', 'E']})
        m = basico.get_reaction_mapping('R3')
        self.assertListEqual(m['product'], ['S', 'E', 'F'])

        # add new reaction with hill kinetics
        basico.add_reaction('R4', 'S -> E', function='Hill Cooperativity')
        basico.add_parameter('hillcoeff')
        basico.add_parameter('K_ER_1')
        basico.add_parameter('v_ER_1')
        basico.add_species('Ca_1')
            
        m1 = basico.get_reaction_mapping('R4')
        basico.set_reaction_mapping('R4', {'h': 2.0, 'substrate': 'Ca_1', 'Shalve': 'K_ER_1', 'V': 'v_ER_1'})
        m2 = basico.get_reaction_mapping('R4')
        self.assertEqual(m2['substrate'], 'Ca_1')
        self.assertEqual(m2['Shalve'], 'K_ER_1')
        self.assertEqual(m2['V'], 'v_ER_1')
        self.assertEqual(m2['h'], 2.0)
        basico.set_reaction_mapping('R4', {'h': 2.0, 'substrate': 'Ca_1', 'Shalve': 0.1, 'V': 0.1})
        m3 = basico.get_reaction_mapping('R4')
        self.assertEqual(m3['Shalve'], 0.1)
        self.assertEqual(m3['V'], 0.1)
        basico.set_reaction_mapping('R4', {'substrate': 'Ca_1', 'Shalve': 'K_ER_1', 'V': 'v_ER_1', 'h': 'hillcoeff'})
        m4 = basico.get_reaction_mapping('R4')
        self.assertEqual(m4['Shalve'], 'K_ER_1')
        self.assertEqual(m4['V'], 'v_ER_1')
        self.assertEqual(m4['h'], 'hillcoeff')


        

    @unittest.skipIf(sys.version_info < (3, 6, 0), 'This test requires assertLogs which is not available before')
    def test_invalid_expression(self):
        self.assertLogs(
            basico.add_parameter(
                'invald', type='assignment',
                expression='Values[offset] .InitialValue + Values[signal] .InitialValue'))

        self.assertLogs(
            basico.add_parameter(
                'invald2',
                initial_expression='Values[offset] .InitialValue + Values[signal] .InitialValue'))

        # check that the interpreter did not crash
        self.assertTrue(True)


class TestBasicoModelConstruction(unittest.TestCase):

    def setUp(self):
        self.dm = basico.new_model(name='Test Model')
        basico.remove_user_defined_functions()

    def tearDown(self):
        basico.remove_datamodel(self.dm)

    def test_units(self):
        basico.set_model_unit(time_unit='s', quantity_unit='mmol', volume_unit='ml',
                              area_unit='cm*cm', length_unit='cm')
        units = basico.get_model_units()
        self.assertEqual(units['time_unit'], 's')
        self.assertEqual(basico.get_time_unit(), 's')
        self.assertEqual(units['quantity_unit'], 'mmol')
        self.assertEqual(units['volume_unit'], 'ml')

    def test_species(self):
        basico.add_species('A', initial_concentration=10)

        # set and get notes
        basico.set_notes('Notes on the species A', name='A')
        notes = basico.get_notes(name='A')
        self.assertTrue('Notes on the species A' in notes)

        a = basico.as_dict(basico.get_species('A'))
        self.assertIsNotNone(a)
        self.assertAlmostEqual(a['initial_concentration'], 10)
        basico.set_species('A', concentration=5)
        a = basico.as_dict(basico.get_species('A'))
        self.assertAlmostEqual(a['concentration'], 5.0)
        basico.set_species('A', exact=True, initial_expression=' 1 / {Time}', expression='{Time}', status='assignment')
        a = basico.as_dict(basico.get_species('A'))
        self.assertEqual(a['type'], 'assignment')
        self.assertEqual(a['expression'], 'Time')
        self.assertEqual(a['initial_expression'], '')

        basico.set_element_name('A', new_name='A0')

        basico.remove_species('A0')
        self.assertIsNone(basico.get_species('A0', exact=True))

    def test_annotation(self):
        basico.add_species('A', initial_concentration=10)
        basico.set_miriam_annotation(name='A', descriptions=[{'id': 'P09560',
   'qualifier': 'is version of',
   'uri': 'http://identifiers.org/uniprot/P09560',
   'resource': 'UniProt Knowledgebase'}])

        annot = basico.get_miriam_annotation(name='A')
        if 'unknown' in annot['descriptions'][0]['id']:
            # this means that annotations have not yet been initialized on the machine
            # this is expected to happen on CI
            self.assertTrue('P09560' in annot['descriptions'][0]['id'])
        else:
            self.assertEqual(annot['descriptions'][0]['id'], 'P09560')

    def test_compartment(self):
        basico.add_compartment('v',  initial_size=2)

        # set and get notes
        basico.set_notes('Notes on the compartment v', name='v')
        notes = basico.get_notes(name='v')
        self.assertTrue('Notes on the compartment v' in notes)

        v = basico.get_compartments('v', exact=True)
        self.assertIsNotNone(v)
        basico.set_compartment('v',  size=3)
        v = basico.as_dict(basico.get_compartments('v', exact=True))
        self.assertAlmostEqual(v['size'], 3)
        basico.set_compartment('v', exact=True, initial_expression=' 1 / {Time}',
                               expression='{Time}', status='assignment')
        v = basico.as_dict(basico.get_compartments('v'))
        self.assertEqual(v['type'], 'assignment')
        self.assertEqual(v['expression'], 'Time')
        self.assertEqual(v['initial_expression'], '')
        self.assertEqual(v['dimensionality'], 3)


        # change dimensions to two
        basico.set_compartment('v', exact=True, dimensionality=2)
        v_df = basico.get_compartments('v')
        v = basico.as_dict(v_df)
        self.assertEqual(v['dimensionality'], 2)

        basico.set_compartment('v', exact=True, dimensionality=v['dimensionality'])
        basico.set_compartment('v', exact=True, dimensionality=v_df.iloc[0]['dimensionality'])

        # apply all recorded changes
        basico.set_compartment(exact=True, **v)
        

        basico.remove_compartment('v')
        self.assertIsNone(basico.get_compartments('v', exact=True))


    def test_global_parameters(self):
        basico.add_parameter('p', initial_value=3)
        v = basico.get_parameters('p', exact=True)
        self.assertIsNotNone(v)
        basico.set_parameters('p', exact=True, initial_expression=' 1 / {Time}', expression='{Time}',
                              status='assignment')
        v = basico.as_dict(basico.get_parameters('p'))
        self.assertEqual(v['type'], 'assignment')
        self.assertEqual(v['expression'], 'Time')
        self.assertEqual(v['initial_expression'], '')
        basico.remove_parameter('p')
        self.assertIsNone(basico.get_parameters('v', exact=True))

    def test_functions(self):
        basico.add_function('fun', infix='v * S / (k + S)', mapping={'S': 'substrate'})

        # set and get notes
        basico.set_notes('Notes on the function fun', name='fun')
        notes = basico.get_notes(name='fun')
        self.assertTrue('Notes on the function fun' in notes)

        fun = basico.as_dict(basico.get_functions('fun', exact=True))
        self.assertIsNotNone(fun)
        self.assertEqual(fun['reversible'], False)
        self.assertEqual(fun['formula'], 'v * S / (k + S)')
        self.assertEqual(fun['mapping']['S'], 'substrate')
        basico.remove_function('fun')
        fun = basico.as_dict(basico.get_functions('fun', exact=True))
        self.assertIsNone(fun)

    def test_events(self):
        basico.add_parameter('p0', initial_value=10)
        basico.add_event('e0', trigger='Time > 10', assignments=[('Values[p0]', 1)],
                         fire_at_initial_time=True, persistent=True, delay_calculation=False)
        
        # set and get notes
        basico.set_notes('Notes on the event e0', name='e0')
        notes = basico.get_notes(name='e0')
        self.assertTrue('Notes on the event e0' in notes)

        e = basico.as_dict(basico.get_events('e0', exact=True))
        self.assertIsNotNone(e)
        self.assertEqual(e['trigger'], 'Time > 10')
        self.assertEqual(e['assignments'][0]['target'], 'Values[p0]')
        self.assertEqual(e['assignments'][0]['expression'], '1')
        self.assertEqual(e['fire_at_initial_time'], True)
        self.assertEqual(e['persistent'], True)
        self.assertEqual(e['delay_calculation'], False)
        basico.remove_event('e0', exact=True)
        e = basico.as_dict(basico.get_events('e0', exact=True))
        self.assertIsNone(e)

    def test_reactions(self):
        basico.add_reaction('r0', 'A -> B')

        # set and get notes
        basico.set_notes('Notes on the reaction r0', name='r0')
        notes = basico.get_notes(name='r0')
        self.assertTrue('Notes on the reaction r0' in notes)

        r0 = basico.as_dict(basico.get_reactions('r0', exact=True))
        self.assertIsNotNone(r0)
        self.assertEqual(r0['scheme'], 'A -> B')
        self.assertEqual(r0['function'], 'Mass action (irreversible)')
        self.assertEqual(r0['mapping']['substrate'], 'A')

        others = basico.as_dict(basico.get_functions(suitable_for='r0'))
        self.assertIsNotNone(others)
        constant = [x for x in others if 'Constant flux' in x['name']]
        self.assertIsNotNone(constant)

        basico.remove_reaction('r0')
        r0 = basico.as_dict(basico.get_reactions('r0', exact=True))
        self.assertIsNone(r0)

    def test_mapping_volumes(self):
        basico.add_function(name='Uni-molecular transport', type='irreversible',
                     infix='Vol*k1*S',
                     mapping={'Vol': 'volume', 'k1': 'parameter', 'S':
                         'substrate'})

        i = 0
        j = 0
        compname = 'cell[{},{}]'.format(i, j)
        basico.add_compartment(name=compname)
        app = '_{},{}'.format(i, j)
        vol1 = compname
        ngb = app
        basico.add_parameter('T0.r_MxferWG')
        basico.add_reaction(name=f'R31_1{app}', scheme=f'EWG1{app} -> EWG4{ngb}', function='Uni-molecular transport',
                     mapping={'Vol': vol1, 'k1':
                         'T0.r_MxferWG'})
        data = basico.as_dict(basico.get_reactions(name=f'R31_1{app}'))
        self.assertEqual(data['function'], 'Uni-molecular transport')
        self.assertDictEqual(data['mapping'], {'Vol': 'cell[0,0]', 'k1': 'T0.r_MxferWG', 'S': 'EWG1_0,0'})

        # ensure this does not crash
        vol1 = 'Compartments[cell[{},{}]].Volume'.format(i, j)
        basico.add_reaction(name=f'R32_1{app}', scheme=f'EWG1{app} -> EWG4{ngb}', function='Uni-molecular transport',
                            mapping={'Vol': vol1, 'k1':
                                'T0.r_MxferWG'})

    def test_parameter_sets(self):
        dm = basico.load_example('LM')
        sol = basico.run_parameter_estimation(method=basico.PE.CURRENT_SOLUTION, create_parametersets=True)
        psets = basico.get_parameter_sets()
        self.assertEqual(len(psets), 6)
        psets = basico.get_parameter_sets(name='Original')
        self.assertEqual(len(psets), 1)

        # remove a specific parameter set
        basico.remove_parameter_sets(name='Original')
        self.assertEqual(len(basico.get_parameter_sets()), 5)

        # add parameter set from model
        basico.add_parameter_set(name='Current State')
        current_state = basico.get_parameter_sets(name='Current State', exact=True);
        self.assertEqual(len(current_state), 1)


        # remove all parameter sets
        basico.remove_parameter_sets()
        self.assertEqual(len(basico.get_parameter_sets()), 0)

        # add parameter set from dictionary
        basico.add_parameter_set(name='Original', param_set_dict=psets[0])

        new_set = basico.get_parameter_sets()
        self.assertEqual(len(new_set), 1)
        self.assertDictEqual(psets[0]['Initial Time'], new_set[0]['Initial Time'])
        self.assertDictEqual(psets[0]['Initial Compartment Sizes'], new_set[0]['Initial Compartment Sizes'])
        self.assertDictEqual(psets[0]['Initial Species Values'], new_set[0]['Initial Species Values'])
        self.assertDictEqual(psets[0]['Initial Global Quantities'], new_set[0]['Initial Global Quantities'])
        self.assertDictEqual(psets[0]['Kinetic Parameters'], new_set[0]['Kinetic Parameters'])

        # apply a parameter set to model
        basico.apply_parameter_set(name='Original')

        # update parameter set from model
        basico.update_parameter_set(name='Original')

        # manually set parameter set to dictionary
        basico.set_parameter_set(name='Original', param_set_dict=psets[0])
        new_set = basico.get_parameter_sets()
        self.assertEqual(len(new_set), 1)

        self.assertDictEqual(psets[0]['Initial Time'], new_set[0]['Initial Time'])
        self.assertDictEqual(psets[0]['Initial Compartment Sizes'], new_set[0]['Initial Compartment Sizes'])
        self.assertDictEqual(psets[0]['Initial Species Values'], new_set[0]['Initial Species Values'])
        self.assertDictEqual(psets[0]['Initial Global Quantities'], new_set[0]['Initial Global Quantities'])
        self.assertDictEqual(psets[0]['Kinetic Parameters'], new_set[0]['Kinetic Parameters'])

        simple_set = basico.get_parameter_sets(values_only=True)
        basico.set_parameter_set(name='Original',param_set_dict=simple_set[0])

        # now test whether the changes actually work
        basico.remove_parameter_sets()
        basico.add_parameter_set(name='Current State')
        basico.add_parameter_set(name='ToModify')

        # try a partial set
        basico.add_parameter_set(name='Partial',
                                 param_set_dict={
                                     'Initial Compartment Sizes': {'compartment': 1},
                                     'Initial Species Values': {'S': 2, 'E': 1}})

        # try changing just the initial concentration os some species
        basico.set_parameter_set(name='ToModify',
                                 param_set_dict={'Initial Species Values': {'S': 5, 'E': 0.1, 'nonexisting': 2}},
                                 remove_others=False)
        simple_set = basico.get_parameter_sets(values_only=True)
        self.assertEqual(simple_set[1]['Initial Species Values']['S'], 5)
        self.assertEqual(simple_set[1]['Initial Species Values']['E'], 0.1)

        # apply the parameter set
        basico.apply_parameter_set(name='ToModify')

        # look at actual model state
        species_list = basico.as_dict(basico.get_species())
        specis_dict = {d['name']: d for d in species_list}
        self.assertEqual(specis_dict['S']['initial_concentration'], 5)

        basico.remove_datamodel(dm)

    def test_retrieval_from_multiple_models(self):
        dm1 = basico.load_example('bruss')
        basico.add_parameter('xy_sum', expression='[X] + [Y]', type='assignment')
        basico.add_parameter('xy_initial_half', expression='Values[xy_sum].InitialValue / 2', type='assignment')
        param1 =  basico.as_dict(basico.get_parameters())
        dm2 = basico.new_model(name='empty')
        param2 = basico.as_dict(basico.get_parameters(model=dm1))
        self.assertEqual(len(param1), len(param2))
        self.assertEqual(param1[0]['expression'], param2[0]['expression'])
        self.assertEqual(param1[1]['expression'], param2[1]['expression'])
        basico.remove_datamodel(dm1)
        basico.remove_datamodel(dm2)

    def test_expresssions(self):
        dm = basico.new_model(name='Test Model');
        basico.add_species('X');
        basico.add_species('Y');
        basico.add_parameter('a', initial_value=1);

        basico.add_parameter('b', initial_value=1);
        basico.set_parameters('b', exact=True, expression='{Values[a].InitialValue}/2', status='assignment')
        params1 = basico.as_dict(basico.get_parameters(type='assignment'))
        basico.set_parameters('b', exact=True, expression='Values[a].InitialValue/2', status='assignment')
        params2 = basico.as_dict(basico.get_parameters(type='assignment'))

        basico.add_parameter('c', initial_value=1);
        basico.set_parameters('c', exact=True, expression='{[X]}+{[Y]}', status='assignment')
        params31 = basico.as_dict(basico.get_parameters(type='assignment'))
        basico.set_parameters('c', exact=True, expression=' {[X]} + {[Y]}', status='assignment')
        params3 = basico.as_dict(basico.get_parameters(type='assignment'))
        basico.set_parameters('c', exact=True, expression=' [X] + [Y]', status='assignment')
        params4 = basico.as_dict(basico.get_parameters(type='assignment'))
        params = basico.as_dict(basico.get_parameters(type='assignment'))


        basico.remove_datamodel(dm);

if __name__ == "__main__":
    unittest.main()



import unittest
import basico


class TestMCA(unittest.TestCase):
    def setUp(self):
        basico.load_example('bruss')
        self.assertIsNotNone(basico.get_species('X', exact=True))

    def test_settings(self):
        settings = basico.get_task_settings(basico.T.MCA)
        self.assertIsNotNone(settings)
        self.assertTrue(settings['method']['Use Reder'])
        self.assertTrue(settings['method']['Use Smallbone'])


    def test_run(self):
        basico.run_mca()
        elasticities = basico.get_elasticities()
        self.assertIsNotNone(elasticities)
        concentration_control_coefficients = basico.get_concentration_control_coefficients()
        self.assertIsNotNone(concentration_control_coefficients)
        flux_control_coefficients = basico.get_flux_control_coefficients()
        self.assertIsNotNone(flux_control_coefficients)



if __name__ == '__main__':
    unittest.main()



"""
Logging
=======

Logging convenience functions.

Originally from the `pypesto` package, see
https://github.com/ICB-DCM/pyPESTO
"""

import logging


def log(
    name: str = 'basico',
    level: int = logging.INFO,
    console: bool = True,
    filename: str = '',
):
    """
    Log messages from `name` with `level` to any combination of console/file.

    Parameters
    ----------
    name:
        The name of the logger.
    level:
        The output level to use.
    console:
        If True, messages are logged to console.
    filename:
        If specified, messages are logged to a file with this name.
    """
    logger = logging.getLogger(name)
    logger.setLevel(level)

    if console:
        ch = logging.StreamHandler()
        ch.setLevel(level)
        logger.addHandler(ch)

    if filename:
        fh = logging.FileHandler(filename)
        fh.setLevel(level)
        logger.addHandler(fh)


def log_to_console(level: int = logging.INFO):
    """
    Log to console.

    Parameters
    ----------
    See the `log` method.
    """
    log(level=level, console=True)


def log_to_file(
    level: int = logging.INFO, filename: str = '.basico_logging.log'
):
    """
    Log to file.

    Parameters
    ----------
    See the `log` method.
    """
    log(level=level, filename=filename)


def log_level_active(logger: logging.Logger, level: int) -> bool:
    """Check whether the requested log level is active in any handler.

    This is useful in case log expressions are costly.

    Parameters
    ----------
    logger:
        The logger.
    level:
        The requested log level.

    Returns
    --------
    active:
        Whether there is a handler registered that handles events of importance
        at least `level` and higher.
    """
    for handler in logger.handlers:
        # it is DEBUG=10, INFO=20, WARNING=30, ERROR=40, CRITICAL=50 increasing
        if handler.level <= level:
            return True
        return False



import logging

import numpy as np
import petab
import petab.simulate
import basico
import pandas as pd
import copasi_petab_importer
import os

logger = logging.getLogger(__name__)


def transform_simulation_df(sim_df, observable_df):
    """ Reverses the transformation of the simulation results

    if there is an observableTransformation is defined for an observable
    that is log or log 10, the according simulation result will be transformed
    back.

    :param sim_df: the simulation data frame
    :type sim_df: pd.DataFrame
    :param observable_df: the observable data frame
    :type observable_df: pd.DataFrame

    :return: the transformed simulation data frame
    :rtype: pd.DataFrame
    """
    result = sim_df.copy(True)
    obs_df = observable_df.reset_index()

    if 'observableTransformation' not in obs_df.columns:
        return result

    if 'simulation' not in result.columns:
        logger.warning('No simulation column found in simulation data frame')
        return result

    if 'observableId' not in obs_df.columns:
        logger.warning('No observableId column found in observable data frame')
        return result

    for _, row in obs_df.iterrows():
        if row['observableTransformation'] == 'log':
            result.loc[result['observableId'] == row['observableId'], 'simulation'] = \
                np.exp(result.loc[result['observableId'] == row['observableId'], 'simulation'])
        elif row['observableTransformation'] == 'log10':
            result.loc[result['observableId'] == row['observableId'], 'simulation'] = \
                np.power(10, result.loc[result['observableId'] == row['observableId'], 'simulation'])

    return result


def create_simulation_df(measurement_df, simulation_results):
    """Creates a simulation data frame

    :param measurement_df: the measurement data frame of the petab problem
    :type measurement_df: pd.DataFrame
    :param simulation_results: the simulation results as obtained from basico.get_simulation_results()
    :type simulation_results: []
    :return: simulation data frame with simulation results in the same format as the measurement data frame
    :rtype: pd.DataFrame
    """
    sim = measurement_df.copy(True)
    sim = sim.rename(columns={"measurement": "simulation"})
    exp_names = basico.get_experiment_names()
    for i in range(len(exp_names)):
        cond_id = exp_names[i]
        s_df = simulation_results[1][i]
        _update_df_from_simulation(sim, s_df, cond_id)
    return sim


def _update_df_from_simulation(petab_df, basico_df, experiment_name):
    """ Updates a simulation data frame with simulation results from one experiment

    :param petab_df: the petab simulation data frame to update
    :type petab_df: pd.DataFrame
    :param basico_df: a basico data frame of simulation results, whose values should be written into `s_df`
    :type basico_df: pd.DataFrame
    :param experiment_name: the experiment (or condition_id) to update in `s_df`
    :type experiment_name: str
    :return: None
    """
    sim_col = petab_df.columns.get_loc('simulation')
    have_time = 'Time' in basico_df.reset_index().columns
    have_no_time_index = 'Time' in basico_df.columns
    if have_time and have_no_time_index:
        basico_df = basico_df.set_index('Time')
    sim_times = basico_df.index.to_list()
    for obs in basico_df.columns.to_list():
        obs_id = obs
        if obs == 'Time':
            continue
        if 'Values[' in obs:
            obs_id = obs[len('Values['):-1]

        values = basico_df[obs].to_list()
        num_values = len(values)

        sub_set = petab_df.loc[
            (petab_df.observableId == obs_id) & (petab_df.simulationConditionId == experiment_name), 'simulation']
        num_expected = len(sub_set)
        if num_expected == 0:
            # this observable is not in the measurements, so we dont have values to update
            continue
        elif num_expected != num_values:
            # we only have certain times and will have to update them individually
            expected = petab_df.loc[
                (petab_df.observableId == obs_id) & (petab_df.simulationConditionId == experiment_name)
                ][['time', 'simulation']]
            for index, row in expected.iterrows():
                if row.time in sim_times and have_time:
                    basic_data = basico_df.loc[row.time, obs]
                    target = petab_df.iloc[index, sim_col]
                    if basic_data.size > 1:
                        # if there are more rows, we still only can take the first one
                        basic_data = basic_data.iloc[0]
                    petab_df.iloc[index, sim_col] = basic_data
                if np.isinf(row.time):
                    # update from steady state:
                    basic_data = basico_df[obs]
                    if basic_data.size > 1:
                        # if there are more rows, we still only can take the first one
                        basic_data = basic_data.iloc[0]
                    petab_df.iloc[index, sim_col] = basic_data
            continue

        petab_df.loc[
            (petab_df.observableId == obs_id) & (petab_df.simulationConditionId == experiment_name), 'simulation'
        ] = values


def petab_llh(pp, sim):
    """Computes the llh based on the simulation data frame

    :param pp: petab problem
    :type pp: petab.Problem

    :param sim: simulation data frame
    :type sim: pd.DataFrame

    :return: the llh as calculated by petab

    """
    return petab.calculate_llh(pp.measurement_df, sim, pp.observable_df, pp.parameter_df)


def petab_chi2(pp, sim):
    """Computes the chi2 based on the simulation data frame

    :param pp: petab problem
    :type pp: petab.Problem

    :param sim: simulation data frame
    :type sim: pd.DataFrame

    :return: the chi2 as calculated by petab

    """
    return petab.calculate_chi2(pp.measurement_df, sim, pp.observable_df, pp.parameter_df)


def petab_residuals(pp, sim):
    """Computes the residuals based on the simulation data frame

    :param pp: petab problem
    :type pp: petab.Problem

    :param sim: simulation data frame
    :type sim: pd.DataFrame

    :return: the residuals as calculated by petab

    """
    return petab.calculate_residuals(pp.measurement_df, sim, pp.observable_df, pp.parameter_df)


def write_problem_to(problem, output_dir, model_name):
    """Writes the given petab problem to the output directory with the given model_name

    :param problem: the petab problem
    :type problem: petab.Problem
    :param output_dir: the output directory where the files should be written to
    :type output_dir: str
    :param model_name: the common model name that all files will have
    :type model_name: str
    :return: dictionary of all the files created
    :rtype: {}
    """
    yaml_file = os.path.join(output_dir, 'problem_{0}.yaml'.format(model_name))
    model_file = os.path.join(output_dir, 'model_{0}.xml'.format(model_name))
    cond_file = os.path.join(output_dir, 'experimentalCondition_{0}.tsv'.format(model_name))
    data_file = os.path.join(output_dir, 'measurementData_{0}.tsv'.format(model_name))
    param_file = os.path.join(output_dir, 'parameters_{0}.tsv'.format(model_name))
    obs_file = os.path.join(output_dir, 'observables_{0}.tsv'.format(model_name))
    files = {
        'model': model_file,
        'conditions': cond_file,
        'measurements': data_file,
        'parameters': param_file,
        'observables': obs_file,
        'problem': yaml_file,
    }
    # write to files
    problem.to_files(
        model_file=model_file,
        condition_file=cond_file,
        measurement_file=data_file,
        parameter_file=param_file,
        observable_file=obs_file,
        yaml_file=yaml_file
    )
    return files


def load_petab(yaml_file, output_dir, out_name=None):
    """Loads the petab file as specified in the yaml file

    This file will be converted using the petab_copasi_importer, and serialized to the specified
    output directory.

    :param yaml_file: the yaml file of the petab description
    :type yaml_file: str
    :param output_dir: the output directory to write the files in
    :type output_dir: str
    :param out_name: optional base name of the copasi file. if it is not specified ith will be
        the basename of the yaml file
    :type out_name: str or None

    :return: the full path to the cps file loded
    """
    # convert to COPASI file
    converter = copasi_petab_importer.PEtabConverter.from_yaml(yaml_file, out_dir=output_dir, out_name=out_name)
    converter.convert()

    # load into basico
    cps_file = os.path.join(output_dir, converter.out_name + '.cps')
    basico.load_model(cps_file)
    return cps_file


class PetabSimulator(petab.simulate.Simulator):
    """ Implementation of a COPASI simulator with the petab simulation interface

        Examples:

        To instantiate the simulatior, just give it the problem and the settings you would like
        to use

        >>> sim = PetabSimulator(pp, settings= {'method' : {'name': PE.EVOLUTIONARY_STRATEGY_SRES}})
        >>> sim.simulate()

    """

    def __init__(self, petab_problem, working_dir=None, settings=None, name='', **kwargs):
        """ Inititializes the simulator

        :param petab_problem: A PEtab problem.
        :type petab_problem: petab.Problem

        :param working_dir: optional working dir (temp directory will be taken if it does not exist
        :type working_dir: Optional[Union[pathlib.Path, str]]

        :param settings: the settings to use for the parameter estimation (by default the converter will
               setup 'Current Solution Statistic', i.e: no parameter estimation will actually be carried out
               and just the values are returned.
        :type settings: {} or None

        :param name: an optional name that will be passed as model specific suffix to use while creating temp
               files (since the petab_problem will be serialized to files in order to be imported)s
        :type name: str

        :param kwargs: will be passed to base class. If it contains an argument:

            * 'cps_file' (str): this COPASI file will be loaded,
        """
        super().__init__(petab_problem, working_dir, **kwargs)
        self.settings = settings
        self.name = name
        self.cps_file = None
        self.model = self.load_model(**kwargs)

    def load_model(self, **kwargs):
        """Loads the model into basico

        If the `cps_file` is in the arguments, then this will be loaded. otherwise this instances
        petab_problem, will be written to files, converted to COPASI and then loaded (at which point
        the cps_file generated will be set).

        :param kwargs:

            - 'cps_file' (str): if present this file will be loaded rather than the petab problem

        :return: None
        """
        if 'cps_file' in kwargs:
            self.cps_file = kwargs['cps_file']
            basico.load_model(self.cps_file)
        else:
            files = write_problem_to(self.petab_problem, self.working_dir, self.name)
            self.cps_file = load_petab(files['problem'], self.working_dir)
        return basico.get_current_model()

    def evaluate(self):
        if self.settings is not None:
            basico.set_task_settings('Parameter Estimation', self.settings)
        return basico.run_parameter_estimation()

    def simulate_without_noise(self):
        """Simulate the PEtab problem.

        This is an abstract method that should be implemented with a simulation
        package. Examples of this are referenced in the class docstring.

        :return:
            Simulated data, as a PEtab measurements table, which should be
            equivalent to replacing all values in the `petab.C.MEASUREMENT`
            column of the measurements table (of the PEtab problem supplied to
            the `__init__` method), with simulated values.

        :rtype: pandas.DataFrame

        """

        simulation_results = basico.get_simulation_results(values_only=True, solution=self.evaluate())
        sim_df = create_simulation_df(self.petab_problem.measurement_df, simulation_results)
        sim_df = basico.petab.transform_simulation_df(sim_df, self.petab_problem.observable_df)
        return sim_df

    def __getstate__(self):
        """Return state for pickling"""
        state = self.__dict__.copy()

        # maybe save cps file to omex? and add to state?
        del state['model']

        return state

    def __setstate__(self, state):
        """Set state after unpickling"""
        self.__dict__.update(state)
        # initialize from 'cps_file' in state, otherwise we'd initialize from the settings
        self.model = self.load_model()


    def set_parameters(self, x_unscaled):
        """Sets the specified fit parameters from the dictionary to the new ones. 

        :param x_unscaled: the new values for the parameters (assumed to be a dictionary 
                           of the form {parameter_id: value}) where the value can be cast to a float
                           and parameter_id is the sbml id of a parameter specified to be estimated
        """
        fit_items = basico.as_dict(basico.get_fit_parameters(model=self.model))
        
        for item in fit_items:
            # remove the Values[] part
            id = item['name'][7:-1] if item['name'].startswith('Values[') else item['name']

            # if we have a new value, set it
            if id in x_unscaled:
                item['start'] = float(x_unscaled[id])

        basico.set_fit_parameters(fit_items, model=self.model)
        return 


"""Submodule with utility methods for carrying out and plotting of parameter estimations.

The main function provided by this submodule is :func:`run_parameter_estimation`. Without any parameters, the
previously set up parameter estimation as stored in the file will be carried out. And the parameters found will
be returned.

It is also possible to set up parameter estimation problems from scratch. To make it as simple as possible, pandas
data frames are used, the mapping from the columns to the model element will be done implicitly by naming the
columns like the corresponding model elements.

Example:

    >>> from basico import *
    >>> m = model_io.load_example("LM-test1")
    >>> print(get_fit_parameters())
    >>> print(get_parameters_solution())
    >>> run_parameter_estimation(method='Levenberg - Marquardt')
    >>> print(get_parameters_solution())

"""
import shutil
import sys

import pandas
import COPASI
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import os
import logging
import yaml

import basico

from . import model_info
from basico.callbacks import get_default_handler

logger = logging.getLogger(__name__)
AFFECTED_EXPERIMENTS = 'Affected Experiments'
TASK_PARAMETER_ESTIMATION = 'Parameter Estimation'


class PE:
    """Constants for Parameter estimation method names

    Convert between method names to enums

        >>> PE.from_enum(0)
        'Current Solution Statistics'

        >>> PE.to_enum('Current Solution Statistics')
        17


    """
    CURRENT_SOLUTION = "Current Solution Statistics"
    RANDOM_SEARCH = "Random Search"
    SIMULATED_ANNEALING = "Simulated Annealing"
    DIFFERENTIAL_EVOLUTION = "Differential Evolution"
    SCATTER_SEARCH = "Scatter Search"
    GENETIC_ALGORITHM = "Genetic Algorithm"
    EVOLUTIONARY_PROGRAMMING = "Evolutionary Programming"
    GENETIC_ALGORITHM_SR = "Genetic Algorithm SR"
    EVOLUTIONARY_STRATEGY_SRES = "Evolution Strategy (SRES)"
    PARTICLE_SWARM = "Particle Swarm"
    LEVENBERG_MARQUARDT = "Levenberg - Marquardt"
    HOOKE_JEEVES = "Hooke & Jeeves"
    NELDER_MEAD = "Nelder - Mead"
    STEEPEST_DESCENT = "Steepest Descent"
    NL2SOL = "NL2SOL"
    PRAXIS = "Praxis"
    TRUNCATED_NEWTON = "Truncated Newton"

    _names = None
    _values = None

    @classmethod
    def _create_name_map(cls):
        return {
            COPASI.CTaskEnum.Method_Statistics: PE.CURRENT_SOLUTION,
            COPASI.CTaskEnum.Method_RandomSearch: PE.RANDOM_SEARCH,
            COPASI.CTaskEnum.Method_SimulatedAnnealing: PE.SIMULATED_ANNEALING,
            COPASI.CTaskEnum.Method_DifferentialEvolution: PE.DIFFERENTIAL_EVOLUTION,
            COPASI.CTaskEnum.Method_ScatterSearch: PE.SCATTER_SEARCH,
            COPASI.CTaskEnum.Method_GeneticAlgorithm: PE.GENETIC_ALGORITHM,
            COPASI.CTaskEnum.Method_GeneticAlgorithmSR: PE.GENETIC_ALGORITHM_SR,
            COPASI.CTaskEnum.Method_SRES: PE.EVOLUTIONARY_STRATEGY_SRES,
            COPASI.CTaskEnum.Method_ParticleSwarm: PE.PARTICLE_SWARM,
            COPASI.CTaskEnum.Method_LevenbergMarquardt: PE.LEVENBERG_MARQUARDT,
            COPASI.CTaskEnum.Method_HookeJeeves: PE.HOOKE_JEEVES,
            COPASI.CTaskEnum.Method_NelderMead: PE.NELDER_MEAD,
            COPASI.CTaskEnum.Method_SteepestDescent: PE.STEEPEST_DESCENT,
            COPASI.CTaskEnum.Method_NL2SOL: PE.NL2SOL,
            COPASI.CTaskEnum.Method_Praxis: PE.PRAXIS,
            COPASI.CTaskEnum.Method_TruncatedNewton: PE.TRUNCATED_NEWTON,
        }

    @classmethod
    def _create_value_map(cls):
        return {
            PE.CURRENT_SOLUTION: COPASI.CTaskEnum.Method_Statistics,
            PE.RANDOM_SEARCH: COPASI.CTaskEnum.Method_RandomSearch,
            PE.SIMULATED_ANNEALING: COPASI.CTaskEnum.Method_SimulatedAnnealing,
            PE.DIFFERENTIAL_EVOLUTION: COPASI.CTaskEnum.Method_DifferentialEvolution,
            PE.SCATTER_SEARCH: COPASI.CTaskEnum.Method_ScatterSearch,
            PE.GENETIC_ALGORITHM: COPASI.CTaskEnum.Method_GeneticAlgorithm,
            PE.GENETIC_ALGORITHM_SR: COPASI.CTaskEnum.Method_GeneticAlgorithmSR,
            PE.EVOLUTIONARY_STRATEGY_SRES: COPASI.CTaskEnum.Method_SRES,
            PE.PARTICLE_SWARM: COPASI.CTaskEnum.Method_ParticleSwarm,
            PE.LEVENBERG_MARQUARDT: COPASI.CTaskEnum.Method_LevenbergMarquardt,
            PE.HOOKE_JEEVES: COPASI.CTaskEnum.Method_HookeJeeves,
            PE.NELDER_MEAD: COPASI.CTaskEnum.Method_NelderMead,
            PE.STEEPEST_DESCENT: COPASI.CTaskEnum.Method_SteepestDescent,
            PE.NL2SOL: COPASI.CTaskEnum.Method_NL2SOL,
            PE.PRAXIS: COPASI.CTaskEnum.Method_Praxis,
            PE.TRUNCATED_NEWTON: COPASI.CTaskEnum.Method_TruncatedNewton,
        }

    @classmethod
    def from_enum(cls, int_value):
        if cls._names is None:
            cls._names = cls._create_name_map()

        return cls._names.get(int_value, PE.CURRENT_SOLUTION)

    @classmethod
    def to_enum(cls, value):
        if cls._values is None:
            cls._values = cls._create_value_map()

        return cls._values.get(value, COPASI.CTaskEnum.Method_Statistics)

    @classmethod
    def all_method_names(cls):
        if cls._names is None:
            cls._names = cls._create_name_map()

        return list(cls._names.values())


try:
    from . import model_io
except ValueError:
    import model_io

try:
    from builtins import ValueError
except ImportError:
    pass


def num_experiment_files(**kwargs):
    """Return the number of experiment files defined.

    :param kwargs:

    - | `model`: to specify the data model to be used (if not specified
      | the one from :func:`.get_current_model` will be taken)

    :return: number of experiment files
    :rtype: int
    """
    model = model_io.get_model_from_dict_or_default(kwargs)
    assert (isinstance(model, COPASI.CDataModel))

    task = model.getTask(TASK_PARAMETER_ESTIMATION)
    assert (isinstance(task, COPASI.CFitTask))

    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CFitProblem))

    return problem.getExperimentSet().size()


def get_experiment_names(**kwargs):
    """Returns the list of experiment names

    :param kwargs:

    - | `model`: to specify the data model to be used (if not specified
      | the one from :func:`.get_current_model` will be taken)

    :return: list of experiment names defined
    :rtype: [str]
    """
    model = model_io.get_model_from_dict_or_default(kwargs)
    assert (isinstance(model, COPASI.CDataModel))

    task = model.getTask(TASK_PARAMETER_ESTIMATION)
    assert (isinstance(task, COPASI.CFitTask))

    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CFitProblem))

    result = []
    for i in range(problem.getExperimentSet().size()):
        experiment = problem.getExperimentSet().getExperiment(i)
        result.append(experiment.getObjectName())
    return result


def _get_experiment_keys(**kwargs):
    model = model_io.get_model_from_dict_or_default(kwargs)
    assert (isinstance(model, COPASI.CDataModel))

    task = model.getTask(TASK_PARAMETER_ESTIMATION)
    assert (isinstance(task, COPASI.CFitTask))

    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CFitProblem))

    result = []
    for i in range(problem.getExperimentSet().size()):
        experiment = problem.getExperimentSet().getExperiment(i)
        result.append(experiment.getKey())
    return result


def num_validations_files(**kwargs):
    """Returns the number of cross validation experiment files

    :param kwargs:

    - | `model`: to specify the data model to be used (if not specified
      | the one from :func:`.get_current_model` will be taken)

    :return: number of cross validation experiment files
    :rtype: int
    """
    model = model_io.get_model_from_dict_or_default(kwargs)
    assert (isinstance(model, COPASI.CDataModel))

    task = model.getTask(TASK_PARAMETER_ESTIMATION)
    assert (isinstance(task, COPASI.CFitTask))

    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CFitProblem))

    return problem.getCrossValidationSet().size()


def _role_to_string(role):
    names = {
        COPASI.CExperiment.time: 'time',
        COPASI.CExperiment.ignore: 'ignored',
        COPASI.CExperiment.independent: 'independent',
        COPASI.CExperiment.dependent: 'dependent',
    }
    return names.get(role, 'ignored')


def _role_to_int(role):
    values = {
        'time': COPASI.CExperiment.time,
        'ignored': COPASI.CExperiment.ignore,
        'independent': COPASI.CExperiment.independent,
        'dependent': COPASI.CExperiment.dependent,
    }
    return values.get(role, COPASI.CExperiment.ignore)


def get_experiment(experiment, **kwargs):
    """Returns the specified experiment.

    :param experiment: experiment name or index
    :type experiment: int or str or COPASI.CExperiment

    :param kwargs:

    - | `model`: to specify the data model to be used (if not specified
      | the one from :func:`.get_current_model` will be taken)

    :return: the experiment or an error if none existent
    """
    if not isinstance(experiment, COPASI.CExperiment):
        model = model_io.get_model_from_dict_or_default(kwargs)
        assert (isinstance(model, COPASI.CDataModel))

        task = model.getTask(TASK_PARAMETER_ESTIMATION)
        assert (isinstance(task, COPASI.CFitTask))

        problem = task.getProblem()
        assert (isinstance(problem, COPASI.CFitProblem))
        exp_set = problem.getExperimentSet()

        if type(experiment) is int and experiment >= exp_set.size():
            raise ValueError('Experiment index out of bounds')
        exp = exp_set.getExperiment(experiment)
        if exp is not None:
            experiment = exp
        else:
            raise ValueError('No experiment for: {0}'.format(experiment))
    return experiment


def _get_experiment_mapping_dict(experiment, **kwargs):
    """ Returns the mapping of the given experiment

    :param experiment: copasi experiment, name or index
    :param kwargs:

    - | `model`: to specify the data model to be used (if not specified
      | the one from :func:`.get_current_model` will be taken)

    :return: list of dictionaries with the mapping for the experiment
    :rtype: [{}]
    """

    experiment = get_experiment(experiment, **kwargs)
    experiment.readColumnNames()
    names = experiment.getColumnNames()
    obj_map = experiment.getObjectMap()
    assert (isinstance(obj_map, COPASI.CExperimentObjectMap))
    rows = []

    last = obj_map.getLastColumn() + 1
    size = obj_map.size()
    max_col = min(len(names), max(last, size))
    for i in range(max_col):
        role = obj_map.getRole(i)
        cn = obj_map.getObjectCN(i)
        cn = model_info._get_cn_string(cn)
        obj = ''
        if cn:
            obj = experiment.getObjectDataModel().getObject(COPASI.CCommonName(cn))
            if obj:
                obj = obj.getObjectDisplayName()

        current = {
            'column': i,
            'type': _role_to_string(role),
            'mapping': obj,
            'cn': cn,
            'column_name': names[i],
        }

        if role == COPASI.CExperiment.dependent:
            scale = obj_map.getScale(i)
            default_scale = obj_map.getDefaultScale(i)
            if scale != default_scale and not np.isnan(scale):
                current['weight'] = scale

        rows.append(current)

    return rows


def get_experiment_mapping(experiment, **kwargs):
    """Retrieves a data frame of the experiment mapping.

    The resulting data frame will have the columns:
    * `column` (int): index of the column in the file
    * `type` (str): 'time', 'dependent', 'indepenent' or 'ignored'
    * 'mapping' (str): the name of the element it is mapped to
    * 'cn' (str): internal identifier

    :param experiment: the experiment to get the mapping from
    :param kwargs:

    - | `model`: to specify the data model to be used (if not specified
      | the one from :func:`.get_current_model` will be taken)

    :return: data frame with the mapping as described
    :rtype: pandas.DataFrame
    """
    rows = _get_experiment_mapping_dict(experiment)
    return pandas.DataFrame(data=rows).set_index('column')


def _get_experiment_file(experiment, **kwargs):
    file_name_only = experiment.getFileNameOnly()
    model = experiment.getObjectDataModel()
    directory = os.path.dirname(model.getFileName())
    return_relative = kwargs.get('return_relative', False)

    if not file_name_only:
        raise ValueError('Invalid Experiment, no filename specified for ' + experiment.getObjectName())

    full_path = os.path.join(directory, os.path.basename(file_name_only))
    if os.path.isfile(full_path):
        if return_relative:
            return os.path.relpath(full_path, directory)
        return full_path

    full_path = os.path.join(directory, file_name_only)
    if os.path.isfile(full_path):
        if return_relative:
            return os.path.relpath(full_path, directory)
        return full_path

    if os.path.isfile(file_name_only):
        if return_relative and directory:
            return os.path.relpath(file_name_only, directory)
        return file_name_only

    file_name = experiment.getFileName()
    if os.path.exists(file_name):
        if return_relative and directory:
            return os.path.relpath(file_name, directory)
        return file_name

    raise_error = kwargs.get('raise_error', True)
    if raise_error:
        raise ValueError('Experiment file {0} does not exist'.format(file_name_only))

    if return_relative and directory and os.path.exists(file_name_only):
        try:
            return os.path.relpath(file_name_only, directory)
        except ValueError:
            # if we can't create a relative path, we copy the file over and then return the relative path
            dst = os.path.join(directory, os.path.basename(file_name_only))
            shutil.copy(file_name_only, dst)
            return os.path.relpath(dst, directory)

    return file_name_only


def get_data_from_experiment(experiment, **kwargs):
    """Returns the data of the given experiment as dataframe

    :param experiment: the experiment
    :param kwargs:

    - | `model`: to specify the data model to be used (if not specified
      | the one from :func:`.get_current_model` will be taken)

    - | `rename_headers` (bool): if true (default) the columns of the headers will be renamed
      | with the names of the element it is mapped to. Also all ignored columns will be removed from the
      | dataset

    :return: dataframe with experimental data
    :rtype: pandas.DataFrame
    """
    experiment = get_experiment(experiment, **kwargs)
    experiment_file = _get_experiment_file(experiment)
    header_row = experiment.getHeaderRow()
    original_headers = None
    num_lines = 0
    separator = experiment.getSeparator()
    with open(experiment_file, encoding='utf-8') as f:
        for line in f:
            num_lines += 1
            if num_lines == header_row:
                original_headers = line.strip().split(separator)
                original_headers = dict(enumerate(original_headers))
    have_headers = header_row < num_lines
    skip_idx = [x-1 for x in range(1, num_lines+1) if
                not (experiment.getFirstRow() <= x <= experiment.getLastRow())]

    if 'rename_headers' in kwargs:
        rename_headers = kwargs['rename_headers']
    else:
        rename_headers = True

    if (have_headers and rename_headers) or original_headers != None:
        skip_idx.insert(0, header_row-1)

    drop_cols = []
    headers = {}
    obj_map = experiment.getObjectMap()
    if rename_headers:
        count = 0
        for i in range(obj_map.size()):
            role = obj_map.getRole(i)

            if role == COPASI.CExperiment.time:
                headers[count] = 'Time'
                count += 1

            elif role == COPASI.CExperiment.ignore:
                drop_cols.append(i)
                count += 1

            else:
                cn = obj_map.getObjectCN(i)
                obj = experiment.getObjectDataModel().getObject(COPASI.CCommonName(cn))
                if obj:
                    headers[count] = obj.getObjectDisplayName()
                    count += 1
                else:
                    drop_cols.append(i)

    if rename_headers or not have_headers:
        df = pandas.read_csv(experiment_file,
                             sep=separator,
                             header=None,
                             skiprows=skip_idx)

    elif original_headers is not None and not rename_headers:
        df = pandas.read_csv(experiment_file,
                             sep=separator,
                             header=None,
                             skiprows=skip_idx)
        df.rename(columns=original_headers, inplace=True)
        return df
    else:
        df = pandas.read_csv(experiment_file,
                             sep=separator,
                             skiprows=skip_idx)

    if not rename_headers:
        return df

    all_columns = list(df.columns)
    for i in range(obj_map.size(), len(all_columns)):
        # drop additional columns not mapped
        drop_cols.append(all_columns[i])

    if any(drop_cols):
        df.drop(drop_cols, axis=1, inplace=True, errors='ignore')

    df.rename(columns=headers, inplace=True)

    return df


def get_experiment_data_from_model(model=None):
    """Returns all experimental data from the model

    :param model: the model to get the data from
    :type model: COPASI.CDataModel or None
    :return: list of dataframes with experimental data (with columns renamed and unmapped columns dropped)
    :rtype: [pandas.DataFrame]
    """
    if model is None:
        model = model_io.get_current_model()
    result = []

    task = model.getTask(TASK_PARAMETER_ESTIMATION)
    assert (isinstance(task, COPASI.CFitTask))

    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CFitProblem))

    experiments = problem.getExperimentSet()
    assert (isinstance(experiments, COPASI.CExperimentSet))

    num_experiments = experiments.getExperimentCount()
    if num_experiments == 0:
        return result

    for i in range(num_experiments):
        experiment = experiments.getExperiment(i)
        df = get_data_from_experiment(experiment, rename_headers=True, model=model)
        result.append(df)

    return result


def get_experiment_filenames(model=None):
    """Returns filenames of all experiments

    :param model: the model to get the data from
    :type model: COPASI.CDataModel or None
    :return: list of filenames of experimental data
    :rtype: [str]
    """
    if model is None:
        model = model_io.get_current_model()
    result = []

    task = model.getTask(TASK_PARAMETER_ESTIMATION)
    assert (isinstance(task, COPASI.CFitTask))

    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CFitProblem))

    experiments = problem.getExperimentSet()
    assert (isinstance(experiments, COPASI.CExperimentSet))

    num_experiments = experiments.getExperimentCount()
    if num_experiments == 0:
        return result

    for i in range(num_experiments):
        experiment = experiments.getExperiment(i)
        result.append(_get_experiment_file(experiment))

    return result


def get_fit_item_template(include_local=False, include_global=False, default_lb=0.001, default_ub=1000, model=None):
    """Returns a template list of items to be used for the parameter estimation

    :param include_local: boolean, indicating whether to include local parameters
    :type include_local: bool

    :param include_global:  boolean indicating whether to include global parameters
    :type include_global: bool

    :param default_lb: default lower bound to be used
    :type default_lb: float

    :param default_ub: default upper bound to be used
    :type default_ub: float

    :param model: the model or None
    :type model: COPASI.CDataModel or None

    :return: List of dictionaries, with the local / global parameters in the format needed by:
             :func:`set_fit_parameters`.
    :rtype: [{}]
    """

    if model is None:
        model = model_io.get_current_model()

    result = []

    if include_global:

        for mv in model.getModel().getModelValues():
            if mv.getStatus() == COPASI.CModelEntity.Status_FIXED:
                result.append({
                    'name': mv.getInitialValueReference().getObjectDisplayName(),
                    'lower': default_lb,
                    'upper': default_ub,
                    'start': mv.getInitialValue()
                })

    if include_local:

        from . import model_info
        local_params = model_info.get_reaction_parameters().reset_index()
        if 'name' in local_params:
            for name, local, value in zip(local_params['name'], local_params['type'], local_params['value']):

                if local == 'local':
                    result.append({
                        'name': name,
                        'lower': default_lb,
                        'upper': default_ub,
                        'start': value
                    })

    return result

def get_fit_parameters(model=None):
    """Returns a data frame with all fit parameters

    The resulting dataframe will have the following columns:

    * `name`: the name of the fit parameter
    * `lower`: the lower bound of the parameter
    * `upper`: the upper bound of the parameter
    * `start`: the start value
    * | `affected`: a list of all experiments (names) the fit parameter should apply to. If empty the parameter should
      | be varied for all experiments.
    * `cn`: internal identifier

    :param model: the model to get the fit parameters from
    :type model: COPASI.CDataModel or None

    :return: data frame with the fit parameters
    :rtype: pandas.DataFrame
    """
    if model is None:
        model = model_io.get_current_model()

    pe_task = model.getTask(TASK_PARAMETER_ESTIMATION)
    problem = pe_task.getProblem()
    assert (isinstance(problem, COPASI.CFitProblem))
    items = problem.getOptItemList()
    data = []

    for i in range(len(items)):
        item = items[i]
        data.append(_get_opt_item_dict(item, model))

    if not data:
        return None

    return pandas.DataFrame(data=data).set_index('name')

def _get_opt_item_dict(item, model):
    """Converts an opt item / constraint to a dictionary

    :param item: the item to convert

    :param model: the model to use

    :return: dictionary with the item data
    """
    obj = model.getObject(COPASI.CCommonName(item.getObjectCN())).toObject().getObjectParent()
    name = obj.getObjectDisplayName()
    item_dict = {
            'name': name,
            'lower': model_info._get_cn_string(item.getLowerBound()),
            'upper': model_info._get_cn_string(item.getUpperBound()),
            'start': item.getStartValue(),
            'affected': _get_affected_experiments(item),
            'cn': model_info._get_cn_string(item.getObjectCN()),
        }
    
    return item_dict

def get_fit_constraints(model=None):
    """Returns a data frame with all fit constraints

    The resulting dataframe will have the following columns:

    * `name`: the name of the fit parameter
    * `lower`: the lower bound of the parameter
    * `upper`: the upper bound of the parameter
    * `start`: the start value
    * | `affected`: a list of all experiments (names) the fit parameter should apply to. If empty the parameter should
      | be varied for all experiments.
    * `cn`: internal identifier

    :param model: the model to get the fit parameters from
    :type model: COPASI.CDataModel or None

    :return: data frame with the fit parameters
    :rtype: pandas.DataFrame
    """
    if model is None:
        model = model_io.get_current_model()

    pe_task = model.getTask(TASK_PARAMETER_ESTIMATION)
    problem = pe_task.getProblem()
    assert (isinstance(problem, COPASI.CFitProblem))

    data = []

    for i in range(problem.getOptConstraintSize()):
        item = problem.getOptConstraint(i).asFitConstraint()
        data.append(_get_opt_item_dict(item, model))

    if not data:
        return None

    return pandas.DataFrame(data=data).set_index('name')


def set_fit_parameters(fit_parameters, model=None):
    """Replaces all existing fit items with the ones provided

    :param fit_parameters: the fit parameters as pandas data frame of list of dictionaries with keys:

           * 'name' str: the display name of the model element to map the column to.
           * 'lower': the lower bound of the parameter
           * 'upper': the upper bound of the parameter
           * 'start' (float, optional): the start value
           * 'affected' (list[str], optional): a list of affected experiment names.
           * 'cn' (str, optional): internal identifier

    :type fit_parameters: pandas.DataFrame or [{}]
    :param model: the model or None
    :type model: COPASI.CDataModel or None
    :return: None
    """
    # type: (pandas.DataFrame, COPASI.CDataModel)
    if model is None:
        model = model_io.get_current_model()

    if type(fit_parameters) is list:
        fit_parameters = pandas.DataFrame(data=fit_parameters)

    pe_task = model.getTask(TASK_PARAMETER_ESTIMATION)
    problem = pe_task.getProblem()
    assert (isinstance(problem, COPASI.CFitProblem))
    while problem.getOptItemSize() > 0:
        problem.removeOptItem(0)

    experiment_keys = _get_experiment_keys(model=model)
    experiment_names = get_experiment_names(model=model)

    if fit_parameters is None:
        return

    for i in range(len(fit_parameters)):
        item = fit_parameters.iloc[i]
        cn = None
        name = None

        if 'cn' in item:
            cn_string = model_info._get_cn_string(item.cn)

            if cn_string:
                cn = COPASI.CRegisteredCommonName(cn_string)

        if 'name' in item and cn is None:
            name = item['name']
            if not cn:
                obj = basico.model_info._get_object(name, initial=True, model=model)
                if obj:
                    cn = obj.getCN()

        if not cn:
            logger.warning('object {0} not found'.format(name))
            continue

        fit_item = problem.addFitItem(cn)
        assert (isinstance(fit_item, COPASI.CFitItem))
        _set_fit_item(experiment_keys, experiment_names, item, fit_item)

def _set_fit_item(experiment_keys, experiment_names, item, fit_item):
    if 'lower' in item:
        fit_item.setLowerBound(COPASI.CRegisteredCommonName(str(item['lower'])))
    if 'upper' in item:
        fit_item.setUpperBound(COPASI.CRegisteredCommonName(str(item['upper'])))
    if 'start' in item:
        fit_item.setStartValue(float(item['start']))
    if 'affected' in item:
        affected = item['affected']

        if type(affected) is float and np.isnan(affected):
            affected = []

        if type(affected) is str:
            affected = [affected]

        if type(affected) is not list:
            logger.warning('affected should be a list or str')
            return
    
        for name in affected:
            if not name:
                continue

            if name not in experiment_names:
                logger.warning('Invalid affected experiment name {0}'.format(name))
                continue

            index = experiment_names.index(name)
            fit_item.addExperiment(experiment_keys[index])


def set_fit_constraints(fit_constraints, model=None):
    """Replaces all existing fit constraints with the ones provided

    :param fit_constraints: the fit parameters as pandas data frame of list of dictionaries with keys:

           * 'name' str: the display name of the model element to map the column to.
           * 'lower': the lower bound of the parameter
           * 'upper': the upper bound of the parameter
           * 'start' (float, optional): the start value
           * 'affected' (list[str], optional): a list of affected experiment names.
           * 'cn' (str, optional): internal identifier

    :type fit_constraints: pandas.DataFrame or [{}]
    :param model: the model or None
    :type model: COPASI.CDataModel or None
    :return: None
    """
    # type: (pandas.DataFrame, COPASI.CDataModel)
    if model is None:
        model = model_io.get_current_model()

    if type(fit_constraints) is list:
        fit_constraints = pandas.DataFrame(data=fit_constraints)

    pe_task = model.getTask(TASK_PARAMETER_ESTIMATION)
    problem = pe_task.getProblem()
    assert (isinstance(problem, COPASI.CFitProblem))

    while problem.getOptConstraintSize() > 0:
        problem.removeOptConstraint(0)

    experiment_keys = _get_experiment_keys(model=model)
    experiment_names = get_experiment_names(model=model)

    if fit_constraints is None:
        return

    for i in range(len(fit_constraints)):
        item = fit_constraints.iloc[i]
        cn = None
        name = None

        if 'cn' in item:
            cn = COPASI.CCommonName(item.cn)

        elif 'name' in item:
            name = item['name']
            if not cn:
                obj = basico.model_info._get_object(name, initial=False, model=model)
                if obj:
                    cn = obj.getCN()

        if not cn:
            logger.warning('object {0} not found'.format(name))
            continue

        fit_item = problem.addFitConstraint(cn)
        assert (isinstance(fit_item, COPASI.CFitConstraint))
        _set_fit_item(experiment_keys, experiment_names, item, fit_item)


def _get_name_for_key(key):
    factory = COPASI.CRootContainer.getKeyFactory()
    obj = factory.get(key)
    if not obj:
        return ''
    return obj.getObjectName()


def _get_affected_experiments(optitem):
    # type: (COPASI.CCopasiParameterGroup) -> [str]
    result = []
    affected = optitem.getGroup(AFFECTED_EXPERIMENTS)
    assert (isinstance(affected, COPASI.CCopasiParameterGroup))
    for i in range(affected.size()):
        current = affected.getParameter(i)
        result.append(_get_name_for_key(current.getStringValue()))
    return result


def get_parameters_solution(model=None):
    """Returns the solution found for the fit parameters as data frame

    The resulting data frame will have the columns:

    * `name`: the name of the parameter
    * `lower`: the parameters lower bound
    * `upper`: the parameters upper bound
    * `sol`: the solution found in the last run (or NaN, if not run yet, or no solution found)
    * `affected`: the experiments this parameter applies to (or an empty list if it applies to all)

    :param model: the model to use, or None
    :type model: COPASI.CDataModel or None
    :return: data frame as described
    :rtype: pandas.DataFrame
    """
    if model is None:
        model = model_io.get_current_model()
    pe_task = model.getTask(TASK_PARAMETER_ESTIMATION)
    problem = pe_task.getProblem()
    assert (isinstance(problem, COPASI.CFitProblem))
    solution = problem.getSolutionVariables()
    items = problem.getOptItemList()
    assert (solution.size() == len(items))
    data = []

    for i in range(solution.size()):
        item = items[i]
        sol = solution.get(i)
        obj = model.getObject(COPASI.CCommonName(item.getObjectCN()))
        if obj is None:
            logger.debug('fit item not in model, cn: {0}'.format(item.getObjectCN()))
            continue
        obj = obj.toObject().getObjectParent()
        name = obj.getObjectDisplayName()
        data.append({
            'name': name,
            'lower': model_info._get_cn_string(item.getLowerBound()),
            'upper': model_info._get_cn_string(item.getUpperBound()),
            'sol': sol,
            'affected': _get_affected_experiments(item),
        })

    if not data:
        return pandas.DataFrame()

    return pandas.DataFrame(data=data).set_index('name')


def _get_role_for_reference(reference_name):
    role_map = {
        'Concentration': COPASI.CExperiment.dependent,
        'ParticleNumber': COPASI.CExperiment.dependent,
        'ParticleNumberRate': COPASI.CExperiment.dependent,
        'InitialConcentration': COPASI.CExperiment.independent,
        'InitialParticleNumber': COPASI.CExperiment.independent,
        'InitialValue': COPASI.CExperiment.independent,
        'InitialVolume': COPASI.CExperiment.independent,
        'Rate': COPASI.CExperiment.dependent,
        'Value': COPASI.CExperiment.dependent,
        'Volume': COPASI.CExperiment.dependent,
    }
    return role_map.get(reference_name, COPASI.CExperiment.ignore)


def add_experiment(name, data, **kwargs):
    """Adds a new experiment to the model.

    This method adds a new experiment file to the parameter estimation task. The provided
    data frame will be written into the current directory as `experiment_name.txt` unless a filename
    has been provided.

    The mapping between the columns and the model elements should be done by having the columns of the data
    frame be model element names in question. So for example `[A]` to note that the transient concentrations
    of a species `A` is to be mapped as dependent variable. or `[A]_0` to note that the initial concentration of
    a species `A` is to be mapped as independent variable.

    :param name: the name of the experiment
    :type name: str
    :param data: the data frame with the experimental data
    :type data: pandas.DataFrame
    :param kwargs:

    - | `model`: to specify the data model to be used (if not specified
      | the one from :func:`.get_current_model` will be taken)

    - | `file_name` (str): the file name to save the experimental data to (otherwise it will be name.txt)

    - | `data_dir` (str): the directory to save the experimental data to (otherwise it will be the current directory)

    :return: the filename of the generated data file
    :rtype: str
    """
    model = model_io.get_model_from_dict_or_default(kwargs)
    assert (isinstance(model, COPASI.CDataModel))

    # compile the model if needed, so that the math model is up to date
    model.getModel().compileIfNecessary()

    task = model.getTask(TASK_PARAMETER_ESTIMATION)
    assert (isinstance(task, COPASI.CFitTask))
    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CFitProblem))
    exp_set = problem.getExperimentSet()
    assert (isinstance(exp_set, COPASI.CExperimentSet))
    exp = exp_set.getExperiment(name)
    if exp is not None:
        logger.error('An experiment with the name {0} already exists'.format(name))
        return None

    # save data as tsv
    data_dir = kwargs.get('data_dir', os.path.curdir)
    file_name = os.path.abspath(os.path.join(data_dir, name.replace(' ', '_') + '.txt'))
    if 'file_name' in kwargs:
        file_name = os.path.abspath(kwargs['file_name'])

    assert (isinstance(data, pd.DataFrame))
    data.to_csv(file_name, sep='\t', header=True, index=False)

    # create experiment
    exp = COPASI.CExperiment(model)
    exp = exp_set.addExperiment(exp)
    info = COPASI.CExperimentFileInfo(exp_set)
    info.setFileName(file_name)
    info.sync()
    exp.setObjectName(name)
    exp.setFileName(file_name)
    exp.setHeaderRow(1)
    exp.setFirstRow(1)
    exp.setLastRow(len(data)+1)

    columns = data.columns.to_list()
    if 'time' in [col.lower() for col in columns]:
        exp.setExperimentType(COPASI.CTaskEnum.Task_timeCourse)
    else:
        exp.setExperimentType(COPASI.CTaskEnum.Task_steadyState)

    obj_map = exp.getObjectMap()
    num_cols = len(columns)
    obj_map.setNumCols(num_cols)

    num_messages = COPASI.CCopasiMessage.size()

    for i in range(num_cols):
        role = COPASI.CExperiment.ignore
        current = columns[i]
        if current.lower() == 'time':
            role = COPASI.CExperiment.time
            obj_map.setRole(i, role)
            continue

        obj = model.findObjectByDisplayName(current)
        if obj is None:
            logger.warning("Can't find model element for {0}".format(current))
        else:
            assert (isinstance(obj, COPASI.CDataObject))
            if obj.getObjectType() != 'Reference':
                try:
                    obj = obj.getValueReference()
                except AttributeError:
                    logger.error("Cannot map the element {0}".format(current))
            role = _get_role_for_reference(obj.getObjectName())
            if not obj_map.setObjectCN(i, obj.getCN()):
                logger.error("Error while mapping data for for {0}".format(current))
        obj_map.setRole(i, role)

    exp.calculateWeights()
    if not exp_set.compile(model.getModel().getMathContainer()):
        logger.error('Failed to add experiment {0}'.format(name))
        logger.error(basico.model_info.get_copasi_messages(num_messages, 'No output'))

    return file_name


def run_parameter_estimation(**kwargs):
    """Runs the parameter estimation task as specified:

    The following are valid methods to be used for the parameter estimation task.

        Current Solution:

            * `Current Solution Statistics`,

        Global Methods:

            * `Random Search`,
            * `Simulated Annealing`,
            * `Differential Evolution`,
            * `Scatter Search`,
            * `Genetic Algorithm`,
            * `Evolutionary Programming`,
            * `Genetic Algorithm SR`,
            * `Evolution Strategy (SRES)`,
            * `Particle Swarm`,

        Local Methods:

            * `Levenberg - Marquardt`,
            * `Hooke & Jeeves`,
            * `Nelder - Mead`,
            * `Steepest Descent`,
            * `NL2SOL`,
            * `Praxis`,
            * `Truncated Newton`,

    :param kwargs:

    - | `model`: to specify the data model to be used (if not specified
      | the one from :func:`.get_current_model` will be taken)

    - | `method` (str): one of the strings from above

    - | `randomize_start_values` (bool): if true, parameters will be randomized before starting otherwise the
      | parameters starting value will be taken.

    - | `calculate_statistics` (bool): if true, the statistics will be calculated at the end of the task

    - | `create_parametersets` (bool): if true, parameter sets will be created for all experiments

    - `use_initial_values` (bool): whether to use initial values

    - `scheduled` (bool): sets whether the task is scheduled or not

    - `update_model` (bool): sets whether the model should be updated, or reset to initial conditions.

    - `settings` (dict): a dictionary with settings to use, in the same format as the ones obtained from
                         :func:`.get_task_settings`

    - `write_report` (bool): overrides the writing of a report file of filename is specified. (defaults to True)

    :return: the solution for the fit parameters see :func:`get_parameters_solution`.
    :rtype: pandas.DataFrame
    """
    model = model_io.get_model_from_dict_or_default(kwargs)
    assert (isinstance(model, COPASI.CDataModel))

    model.getModel().compileIfNecessary()

    task = model.getTask(TASK_PARAMETER_ESTIMATION)
    assert (isinstance(task, COPASI.CFitTask))

    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CFitProblem))

    if problem.getOptItemSize() == 0:
        logger.warning('No fit parameters defined, skipping parameter estimation run')
        return get_parameters_solution(model)

    if 'scheduled' in kwargs:
        task.setScheduled(kwargs['scheduled'])

    if 'update_model' in kwargs:
        task.setUpdateModel(kwargs['update_model'])

    old_create_parameter_sets = problem.getCreateParameterSets()
    # old_calculate_statistics = problem.getCalculateStatistics()
    # old_randomize_start_values = problem.getRandomizeStartValues()

    # problem.setCreateParameterSets(True)

    write_report = kwargs.get('write_report', True)
    report_name = task.getReport().getTarget()
    if not write_report:
        task.getReport().setTarget('')

    if 'method' in kwargs:
        method = kwargs['method']
        if isinstance(method, int):
            task.setMethodType(method)
        else:
            task.setMethodType(COPASI.CCopasiMethod.TypeNameToEnum(method))

    if 'randomize_start_values' in kwargs:
        problem.setRandomizeStartValues(bool(kwargs['randomize_start_values']))

    if 'calculate_statistics' in kwargs:
        problem.setCalculateStatistics(bool(kwargs['calculate_statistics']))

    if 'create_parametersets' in kwargs:
        problem.setCreateParameterSets(bool(kwargs['create_parametersets']))

    use_initial_values = kwargs.get('use_initial_values', True)

    if 'settings' in kwargs:
        basico.set_task_settings(task, kwargs['settings'])

    # the parameter estimation task will not run if errors have not been
    # cleared from the error log. So we clear them here if necessary
    if COPASI.CCopasiMessage.getHighestSeverity() > COPASI.CCopasiMessage.WARNING:
        logger.warning("Uncaptured Errors: " +
        basico.model_info.get_copasi_messages(0, 'No output'))

    num_messages_before = COPASI.CCopasiMessage.size()

    task.setCallBack(get_default_handler())
    result = task.initializeRaw(COPASI.CCopasiTask.OUTPUT_UI)
    if not result:
        logger.error("Error while initializing parameter estimation: " +
        basico.model_info.get_copasi_messages(num_messages_before, 'No output'))
    else:
        result = task.processRaw(use_initial_values)
        if not result:
            logger.error("Error while initializing parameter estimation: " +
            basico.model_info.get_copasi_messages(num_messages_before))

    task.restore()

    problem.setCreateParameterSets(old_create_parameter_sets)

    if not write_report:
        task.getReport().setTarget(report_name)

    return get_parameters_solution(model)


def get_simulation_results(values_only=False, update_parameters=True, **kwargs):
    """Runs the current solution statistics and returns result of simulation and experimental data

    :param values_only: if true, only time points at the measurements will be returned
    :type values_only: bool

    :param update_parameters: if set true, the model will be updated with the parameters
           found from the solution. (defaults to True)
    :type update_parameters: bool

    :param kwargs:

    - | `model`: to specify the data model to be used (if not specified
      | the one from :func:`.get_current_model` will be taken)

    - | `solution`: a solution data frame to use, if not specified a current solution
                    statistic will be computed

    :return: tuple of lists of experiment data, and a list of simulation data
    :rtype: ([pandas.DataFrame],[pandas.DataFrame])
    """
    import basico
    dm = model_io.get_model_from_dict_or_default(kwargs)

    task = dm.getTask(TASK_PARAMETER_ESTIMATION)
    assert (isinstance(task, COPASI.CFitTask))

    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CFitProblem))

    experiments = problem.getExperimentSet()
    assert (isinstance(experiments, COPASI.CExperimentSet))

    result = []
    num_experiments = experiments.getExperimentCount()
    if num_experiments == 0:
        return result

    if update_parameters:
        if 'solution' in kwargs:
            solution = kwargs['solution']
            if type(solution) is dict or type(solution) is list:
                solution = pd.DataFrame(data=solution)
        else:
            settings = basico.get_task_settings(basico.T.PARAMETER_ESTIMATION, model=dm)
            solution = run_parameter_estimation(method='Current Solution Statistics', write_report=False, randomize_start_values=False, model=dm)
            # restore settings
            basico.set_task_settings(basico.T.PARAMETER_ESTIMATION, settings)

    exp_data = []
    sim_data = []

    for i in range(num_experiments):
        experiment = experiments.getExperiment(i)
        exp_name = experiment.getObjectName()
        df = get_data_from_experiment(experiment, rename_headers=True)
        mapping = get_experiment_mapping(experiment)

        is_steady_state = experiment.getExperimentType() == COPASI.CTaskEnum.Task_steadyState
        num_independent_points = df.shape[0]
        steady_state_task = dm.getTask(basico.T.STEADY_STATE)
        container = dm.getModel().getMathContainer()

        if update_parameters:
            _update_fit_parameters_from(dm, solution, exp_name)

        container.fetchInitialState()
        container.updateInitialValues(COPASI.CCore.Framework_ParticleNumbers)
        container.applyInitialValues()
        container.updateSimulatedValues(False)
        container.updateTransientDataValues()
        experiment.updateModelWithIndependentData(0)
        container.pushAllTransientValues()
        container.pushInitialState()

        if is_steady_state:
            # run steady state
            steady_state_task.initializeRaw(COPASI.CCopasiTask.OUTPUT_UI)
            steady_state_task.processRaw(True)
            data = basico.model_info._collect_data(cns=mapping[mapping.type == 'dependent']['cn'].to_list()).transpose()

            for j in range(1, num_independent_points):
                container.applyInitialValues()
                container.updateSimulatedValues(False)
                container.updateTransientDataValues()
                experiment.updateModelWithIndependentData(j)
                container.pushAllTransientValues()
                container.pushInitialState()

                if update_parameters:
                    _update_fit_parameters_from(dm, solution, exp_name)
                steady_state_task.processRaw(True)

                new_row = basico.model_info._collect_data(
                    cns=mapping[mapping.type == 'dependent']['cn'].to_list()).transpose()
                data = pd.concat([data, new_row], ignore_index=True)

        else:
            # run time course (getting only the data from the experiment)
            duration = df.iloc[-1].Time
            cols = ['Time'] + mapping[mapping.type == 'dependent']['cn'].to_list()
            tc_settings = basico.get_task_settings(basico.T.TIME_COURSE, model=dm)
            if values_only:
                data = basico.run_time_course_with_output(output_selection=cols, values=df.Time.to_list(), start_time=df.iloc[0].Time)
            else:
                data = basico.run_time_course_with_output(output_selection=cols,duration=duration)
            # restore setttings
            basico.set_task_settings(basico.T.TIME_COURSE, tc_settings)

        exp_data.append(df)
        sim_data.append(data)

    return exp_data, sim_data


def _apply_nth_change(change, columns, df, dm, exp_name, independent, model, num_independent):
    change_set = COPASI.DataObjectSet()
    for j in range(num_independent):
        name = independent.iloc[j].mapping
        cn = independent.iloc[j].cn

        if name not in columns:
            # independent value is not found in df
            continue

        value = df.iloc[change][name]
        obj = dm.getObject(COPASI.CCommonName(cn))

        if obj is None:  # not found skip
            logger.debug('independent object not found for cn: {0}'.format(cn))
            continue

        if obj.getObjectName() == 'InitialConcentration':
            obj.getObjectParent().setInitialConcentration(value)
        else:
            obj.getObjectParent().setInitialValue(value)

        change_set.append(obj)
        logger.debug('set independent "{0}" to "{1}"'.format(cn, value))
    if change_set.size() > 0:
        model.updateInitialValues(change_set)


def _get_value_from_bound(bound):
    """

    :param bound: the bound for the fit item (a float value as string, or name of another reference)
    :return: the value of the bound
    :rtype: float
    """
    try:
        if isinstance(bound, COPASI.CRegisteredCommonName):
            value = basico.get_value(bound)
            if value is None: 
                value = bound.getString()
        else:
            value = float(bound)
    except ValueError:
        value = basico.get_value(bound)
    return value


def _update_fit_parameters_from(dm, solution, exp_name=''):
    """ Utility function that updates the models fit parameters for the given solution

    :param dm: the current model
    :param solution: the computed solution dataframe
    :param exp_name: the affected experiment or empty for all
    :return: None
    """
    change_set = COPASI.DataObjectSet()
    model = dm.getModel()
    params = get_fit_parameters(dm)
    for j in range(solution.shape[0]):
        name = solution.iloc[j].name
        value = solution.iloc[j].sol
        lower= _get_value_from_bound(solution.iloc[j].lower)
        upper = _get_value_from_bound(solution.iloc[j].upper)

        cn = params.iloc[j].cn
        if np.isnan(value):
            continue
        affected = solution.iloc[j].affected
        if any(affected) and exp_name not in affected:
            continue

        # ensure that values is within the constraint
        if value < lower:
            value = lower
        if value > upper:
            value = upper

        obj = dm.getObject(COPASI.CCommonName(cn))

        if obj is None:  # not found skip
            continue

        if type(obj) is COPASI.CDataObject:

            if obj.getObjectName() == 'InitialConcentration':
                obj.getObjectParent().setInitialConcentration(value)
            elif type(obj.getObjectParent()) is COPASI.CCopasiParameter:
                obj.setDblValue(value)
                model.updateInitialValues(obj)
            else:
                obj.getObjectParent().setInitialValue(value)

            change_set.append(obj)
            logger.debug('set solution value "{0}" to "{1}"'.format(cn, value))
        else:
            basico.set_reaction_parameters(name, value=value)
            logger.debug('set reaction parameter "{0}" to "{1}"'.format(name, value))
    if change_set.size() > 0:
        model.updateInitialValues(change_set)


def plot_per_experiment(**kwargs):
    """
    This function creates one figure per experiment defined, with plots of all dependent variables
    and their fit in it.

    :param kwargs:

    - | `model`: to specify the data model to be used (if not specified
      | the one from :func:`.get_current_model` will be taken)

    :return: array of tuples (fig, ax) for the figures created
    """
    dm = model_io.get_model_from_dict_or_default(kwargs)

    task = dm.getTask(TASK_PARAMETER_ESTIMATION)
    assert (isinstance(task, COPASI.CFitTask))

    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CFitProblem))

    experiments = problem.getExperimentSet()
    assert (isinstance(experiments, COPASI.CExperimentSet))

    result = []
    num_experiments = experiments.getExperimentCount()
    if num_experiments == 0:
        return result

    exp_data, sim_data = get_simulation_results(**kwargs)

    for i in range(num_experiments):
        fig, ax = plt.subplots()
        cycler = plt.cycler("color", plt.cm.tab20c.colors)()
        experiment = experiments.getExperiment(i)
        exp_name = experiment.getObjectName()
        mapping = get_experiment_mapping(experiment)
        ax.set_title(exp_name)

        # set independent values for that experiment
        dependent = mapping[mapping.type == 'dependent']

        num_dependent = dependent.shape[0]
        for j in range(num_dependent):
            nextval = next(cycler)['color']
            name = dependent.iloc[j].mapping
            if name not in sim_data[i].columns:
                name = name[1:-1]
            sim_data[i].reset_index().plot(x='Time', y=name,
                                           label="{0} Fit".format(name), ax=ax, color=nextval)
            name = dependent.iloc[j].mapping
            exp_data[i].plot.scatter(x='Time', y=name, ax=ax, color=nextval,
                                     label='{0} Measured'.format(name))
        result.append((fig, ax))

    return result


def plot_per_dependent_variable(**kwargs):
    """
    This function creates a figure for each dependent variable, with traces for all experiments.

    :param kwargs:

    - | `model`: to specify the data model to be used (if not specified
      | the one from :func:`.get_current_model` will be taken)

    :return: array of tuples (fig, ax) for each figure created
    """
    dm = model_io.get_model_from_dict_or_default(kwargs)

    task = dm.getTask(TASK_PARAMETER_ESTIMATION)
    assert (isinstance(task, COPASI.CFitTask))

    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CFitProblem))

    experiments = problem.getExperimentSet()
    assert (isinstance(experiments, COPASI.CExperimentSet))

    result = []
    num_experiments = experiments.getExperimentCount()
    if num_experiments == 0:
        return result

    exp_data, sim_data = get_simulation_results(**kwargs)

    dependent_variables = {}

    for i in range(num_experiments):
        experiment = experiments.getExperiment(i)
        mapping = get_experiment_mapping(experiment)

        # set independent values for that experiment
        dependent = mapping[mapping.type == 'dependent']
        num_dependent = dependent.shape[0]
        for j in range(num_dependent):
            name = dependent.iloc[j].mapping
            if name not in dependent_variables:
                dependent_variables[name] = []
            dependent_variables[name].append(i)

    for dependent in dependent_variables:
        fig, ax = plt.subplots()
        cycler = plt.cycler("color", plt.cm.tab20c.colors)()
        ax.set_title(dependent)
        experiment_indices = dependent_variables[dependent]

        for i in experiment_indices:
            experiment = experiments.getExperiment(i)
            exp_name = experiment.getObjectName()
            nextval = next(cycler)['color']
            name = dependent
            if name not in sim_data[i].columns:
                name = name[1:-1]

            sim_data[i].reset_index().plot(x='Time', y=name,
                                           label="{0} Fit".format(exp_name), ax=ax, color=nextval)
            exp_data[i].plot.scatter(x='Time', y=dependent, ax=ax, color=nextval,
                                     label='{0} Measured'.format(exp_name))
        result.append((fig, ax))

    return result


def prune_simulation_results(simulation_results):
    """Removes all columns & time points from the simulation set, that are not available in the measurement set

    :param simulation_results: the simulation result as obtained by get_simulation_results

    :return:
    """
    assert len(simulation_results[0]) == len(simulation_results[1])
    for i in range(len(simulation_results[0])):
        s_df = simulation_results[1][i].reset_index()
        e_df = simulation_results[0][i]
        if 'Time' in s_df.columns and 'Time' in e_df.columns:
            s_df = s_df[s_df.Time.isin(e_df.Time.to_list())]
        s_df = s_df.reset_index()
        common_cols = [c for c in e_df.columns.to_list() if c in s_df.columns.to_list()]
        s_df = s_df[common_cols]
        simulation_results[1][i] = s_df

    return simulation_results


def get_fit_statistic(include_parameters=False, include_experiments=False, include_fitted=False, **kwargs):
    """Return information about the last fit.

    :param include_parameters: whether to include information about the parameters in a result entry
                               with key `parameters`
    :type include_parameters: bool

    :param include_experiments: whether to include information about the experiments in a result entry
                               with key `experiments`
    :type include_experiments: bool

    :param include_fitted: whether to include information about the fitted values in a result entry
                               with key `fitted`
    :type include_fitted: bool

    :param kwargs:

    - | `model`: to specify the data model to be used (if not specified
      | the one from :func:`.get_current_model` will be taken)

    :return: dictionary with the fit statistic
    :rtype: {}
    """
    dm = model_io.get_model_from_dict_or_default(kwargs)

    task = dm.getTask(TASK_PARAMETER_ESTIMATION)
    assert (isinstance(task, COPASI.CFitTask))

    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CFitProblem))

    experiments = problem.getExperimentSet()
    assert (isinstance(experiments, COPASI.CExperimentSet))

    function_evaluations = problem.getFunctionEvaluations()
    performed_iterations = function_evaluations > 0
    result = {
        'obj': problem.getSolutionValue(),
        'rms': problem.getRMS(),
        'sd': problem.getStdDeviation(),
        'f_evals': function_evaluations,
        'failed_evals_exception': problem.getFailedEvaluationsExc(),
        'failed_evals_nan': problem.getFailedEvaluationsNaN(),
        'constraint_evals': problem.getConstraintEvaluations(),
        'failed_constraint_evals': problem.geFailedConstraintCounter(),
        'cpu_time': problem.getExecutionTime(),
        'data_points': experiments.getDataPointCount(),
        'valid_data_points': experiments.getValidValueCount(),
    }
    result['evals_per_sec'] = result['cpu_time'] / function_evaluations if performed_iterations else 0

    sol = problem.getSolutionVariables()
    grad = problem.getVariableGradients()
    std = problem.getVariableStdDeviations()

    if include_parameters:
        parameters = []
        for i in range(problem.getOptItemSize()):
            current = problem.getOptItem(i)
            assert (isinstance(current, COPASI.COptItem))
            obj = dm.getObject(current.getObjectCN())
            name = obj.getObjectDisplayName() if obj is not None else 'Not Found'
            parameters.append({
                'name':  name,
                'lower': current.getLowerBound(),
                'start': current.getLastStartValue(),
                'value': sol.get(i) if performed_iterations else np.nan,
                'upper': current.getUpperBound(),
                'std_dev': std.get(i)  if performed_iterations else np.nan,
                'coeff_of_variation': abs(100 * std.get(i) / sol.get(i)) if performed_iterations else np.nan,
                'gradient': grad.get(i) if performed_iterations else np.nan
            })
        result['parameters'] = parameters

    if include_experiments:
        if COPASI.CVersion.VERSION.getVersionDevel() < 263:
            raise ValueError("Newer COPASI version required to return experiment statistic")

        experiment_stats = []
        for i in range(experiments.getExperimentCount()):
            exp = experiments.getExperiment(i)

            valid_value_count = exp.getValidValueCount()
            total_value_count = exp.getTotalValueCount()

            item = {
                'name': exp.getObjectName(),
                'valid_points': valid_value_count,
                'total_points': total_value_count,
                'obj': exp.getObjectiveValue(),
                'rms': exp.getRMS(),
                'error_mean': exp.getErrorMean(),
                'error_mean_sd': exp.getErrorMeanSD(),
                'dependent_data': []
            }

            objects = experiments.getDependentObjects()

            for j in range(objects.size()):
                obj = experiments.getDependentObjects().get(j)
                count = exp.getColumnValidValueCount(obj)
                if not count:
                    continue
                item['dependent_data'].append(
                    {
                        'name': obj.getObjectDisplayName(),
                        'data_points': count,
                        'obj': exp.getObjectiveValue(obj),
                        'rms': exp.getRMS(obj),
                        'error_mean': exp.getErrorSum(obj) / count,
                    }
                )
            experiment_stats.append(item)

        result['experiments'] = experiment_stats

    if include_fitted:
        dependent = experiments.getDependentObjects()
        num_dependent = dependent.size()
        if function_evaluations == 0:
            num_dependent = 0
        values = []
        for i in range(num_dependent):
            current = dependent.get(i)
            name = current.getObjectDisplayName() if current is not None else 'Not Found'
            values.append({
                'name': name,
                'data_points': experiments.getDependentDataCount().get(i),
                'obj': experiments.getDependentObjectiveValues().get(i),
                'rms': experiments.getDependentRMS().get(i),
                'error_mean': experiments.getDependentErrorMean().get(i),
                'error_mean_sd': experiments.getDependentErrorMeanSD().get(i)
            })

        result['variables'] = values

    return result


def remove_experiments(**kwargs):
    """Removes all experiments from the model

    :param kwargs:

    - | `model`: to specify the data model to be used (if not specified
      | the one from :func:`.get_current_model` will be taken)

    :return: None
    """
    dm = model_io.get_model_from_dict_or_default(kwargs)

    task = dm.getTask(TASK_PARAMETER_ESTIMATION)
    assert (isinstance(task, COPASI.CFitTask))

    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CFitProblem))

    experiments = problem.getExperimentSet()
    assert (isinstance(experiments, COPASI.CExperimentSet))

    for _ in range(experiments.size()):
        experiments.removeExperiment(0)


def remove_fit_parameters(**kwargs):
    """Removes all fit items

    :param kwargs:

    - | `model`: to specify the data model to be used (if not specified
      | the one from :func:`.get_current_model` will be taken)

    :return: None
    """
    dm = model_io.get_model_from_dict_or_default(kwargs)

    pe_task = dm.getTask(TASK_PARAMETER_ESTIMATION)
    problem = pe_task.getProblem()
    assert (isinstance(problem, COPASI.CFitProblem))
    while problem.getOptItemSize() > 0:
        problem.removeOptItem(0)

def _weight_method_to_string(weight_method):
    """ Convenience function converting a weight method to string

    :param weight_method: the weight method
    :type weight_method: int
    :return: name of the weight method
    :rtype: str
    """
    weight_map = {
        COPASI.CExperiment.MEAN: 'Mean',
        COPASI.CExperiment.MEAN_SQUARE: 'Mean Square',
        COPASI.CExperiment.SD: 'Standard Deviation',
        COPASI.CExperiment.VALUE_SCALING: 'Value Scaling'
    }
    return weight_map.get(weight_method, 'Mean Square')

def _weight_method_to_int(weight_method):
    """ Convenience function converting a weight method to int

    :param weight_method: the weight method
    :type weight_method: str
    :return: type of the weight method
    :rtype: int
    """
    weight_map = {
        'Mean': COPASI.CExperiment.MEAN,
        'Mean Square': COPASI.CExperiment.MEAN_SQUARE,
        'Standard Deviation': COPASI.CExperiment.SD,
        'Value Scaling': COPASI.CExperiment.VALUE_SCALING
    }
    return weight_map.get(weight_method, COPASI.CExperiment.MEAN_SQUARE)


def get_experiment_dict(experiment, **kwargs):
    """ Returns all information about the experiment as dictionary

    :param experiment: copasi experiment, experiment name or index
    :param kwargs: optional arguments

    - | `model`: to specify the data model to be used (if not specified
      | the one from :func:`.get_current_model` will be taken)

    - | `raise_error`: boolean indicating that an error should be raised if the
      |  experimentfile is not present (default: False)

    - | `return_relative`: to indicate that relative experiment filenames should
      | be returned (default: True)

    :return: all information about the experiment as dictionary
    """
    experiment = get_experiment(experiment, **kwargs)

    kwargs['raise_error'] = kwargs.get('raise_error', False)
    kwargs['return_relative'] = kwargs.get('return_relative', True)
    filename = _get_experiment_file(experiment, **kwargs)

    result = {
        'name': experiment.getObjectName(),
        'filename': filename,
        'type': basico.T.STEADY_STATE if experiment.getExperimentType() == COPASI.CTaskEnum.Task_steadyState else
                basico.T.TIME_COURSE,
        'separator': experiment.getSeparator(),
        'first_row': experiment.getFirstRow(),
        'last_row': experiment.getLastRow(),
        'weight_method': _weight_method_to_string(experiment.getWeightMethod()),
        'normalize_per_experiment': experiment.getNormalizeWeightsPerExperiment()
    }

    if experiment.getHeaderRow() < experiment.getLastRow():
        result['header_row'] = experiment.getHeaderRow()

    mapping = _get_experiment_mapping_dict(experiment, **kwargs)

    # rename / clear up things
    for entry in mapping:
        if 'column_name' in entry:
            if entry['column_name']:
                entry['column'] = entry['column_name']
            del entry['column_name']
        if entry['cn'] == '':
            del entry['cn']
        if entry['mapping'] == '':
            del entry['mapping']

        if 'mapping' in entry:
            entry['object'] = entry['mapping']
            del entry['mapping']

    result['mapping'] = mapping

    return result


def save_experiments_to_dict(**kwargs):
    """Returns a list of dictionaries with the parameter estimation experiments

    :param kwargs: optional arguments

    - | `model`: to specify the data model to be used (if not specified
      | the one from :func:`.get_current_model` will be taken)

    - | `raise_error`: boolean indicating that an error should be raised if the
      |  experimentfile is not present (default: False)

    - | `return_relative`: to indicate that relative experiment filenames should
      | be returned (default: True)

    :return: the parameter estimation experimetns as list of dictionary
    :rtype: [{}]
    """
    experiments = []
    model = model_io.get_model_from_dict_or_default(kwargs)
    assert (isinstance(model, COPASI.CDataModel))

    task = model.getTask(TASK_PARAMETER_ESTIMATION)
    assert (isinstance(task, COPASI.CFitTask))

    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CFitProblem))
    exp_set = problem.getExperimentSet()

    for i in range(exp_set.size()):
        experiments.append(get_experiment_dict(exp_set.getExperiment(i), **kwargs))

    return experiments

def save_experiments_to_yaml(filename=None, **kwargs):
    """Saves the experiment to yaml

    :param filename: optional filename to write to
    :param kwargs: optional arguments

    - | `model`: to specify the data model to be used (if not specified
      | the one from :func:`.get_current_model` will be taken)

    :return: the yaml string
    """
    experiments = save_experiments_to_dict(**kwargs)
    yaml_str = yaml.safe_dump(experiments, indent=2, sort_keys=False, default_flow_style=False )
    if not filename:
         return yaml_str

    with open(filename, 'w', encoding='utf-8') as out_file:
        out_file.write(yaml_str)

    return yaml_str


def load_experiments_from_yaml(experiment_description, **kwargs):
    """ Loads all experiments from the specified experiment description

    All existing experiments will be replaced with the ones from the specified file.

    :param experiment_description: filename or yamlstring
    :param kwargs: optional arguments

    - | `model`: to specify the data model to be used (if not specified
      | the one from :func:`.get_current_model` will be taken)

    :return:
    """

    if os.path.exists(experiment_description):
        with open(experiment_description, 'r') as stream:
            experiments = yaml.safe_load(stream)
    else:
        experiments = yaml.safe_load(experiment_description)

    return load_experiments_from_dict(experiments, **kwargs)


def load_experiments_from_dict(experiments, **kwargs):
    """ Loads all experiments from the specified experiment description

    All existing experiments will be replaced with the ones from the specified file.

    :param experiments: list of experiment dictionaries
    :param kwargs: optional arguments

    - | `model`: to specify the data model to be used (if not specified
      | the one from :func:`.get_current_model` will be taken)

    :return:
    """
    model = model_io.get_model_from_dict_or_default(kwargs)
    fit_items = get_fit_parameters(model)

    # remove existing experiments
    remove_experiments(**kwargs)

    if type(experiments) is dict:
        experiments = [experiments]

    for exp_dict in experiments:
        add_experiment_from_dict(exp_dict, **kwargs)

    # update fit items, as the affected experiments might need updating
    set_fit_parameters(fit_items, model=model)

def _get_nth_line_from_file(filename, n, max_line=None):
    with open(filename, 'r') as stream:
        for i, line in enumerate(stream):
            if i + 1 == n:
                return line.strip()
            if max_line and i == max_line:
                break
    return None


def _get_first_column(mappings, column):
    for entry in mappings:
        if entry['column'] == column:
            return entry
    return None


def add_experiment_from_dict(exp_dict, **kwargs):
    """ Adds an experiment from dictionary

    :param exp_dict:
    :return:
    """
    model = model_io.get_model_from_dict_or_default(kwargs)
    assert (isinstance(model, COPASI.CDataModel))

    task = model.getTask(TASK_PARAMETER_ESTIMATION)
    assert (isinstance(task, COPASI.CFitTask))

    problem = task.getProblem()
    assert (isinstance(problem, COPASI.CFitProblem))

    exp_set = problem.getExperimentSet()
    assert (isinstance(exp_set, COPASI.CExperimentSet))

    data_file_name = exp_dict['filename']

    abs_data_file_name = data_file_name
    if not os.path.isabs(data_file_name):
        abs_data_file_name = os.path.join(os.path.dirname(model.getFileName()), data_file_name)

    if not os.path.exists(abs_data_file_name):
        if 'data_dir' in kwargs:
            abs_data_file_name = os.path.join(kwargs['data_dir'], data_file_name)
        else:
            abs_data_file_name = os.path.abspath(data_file_name)

    if not os.path.exists(abs_data_file_name):
        raise IOError("File not found: " + data_file_name)

    experiment = COPASI.CExperiment(model, exp_dict['name'])
    experiment.setFirstRow(exp_dict['first_row'])
    experiment.setLastRow(exp_dict['last_row'])
    if 'header_row' in exp_dict:
        experiment.setHeaderRow(exp_dict['header_row'])
    experiment.setFileName(data_file_name)
    experiment.setNormalizeWeightsPerExperiment(exp_dict['normalize_per_experiment'])
    experiment.setExperimentType(basico.T.to_enum(exp_dict['type']))
    experiment.setSeparator(exp_dict['separator'])
    experiment.setWeightMethod(_weight_method_to_int(exp_dict['weight_method']))
    columnNumber = experiment.guessColumnNumber()
    experiment.setNumColumns(columnNumber)

    obj_map = experiment.getObjectMap()
    obj_map.setNumCols(columnNumber)

    ## this should not be crashing (needs new COPASI release)
    #
    # experiment.readColumnNames()
    # names = experiment.getColumnNames()
    #
    ## instead we use:

    names = None
    if  'header_row' in exp_dict:
        names = _get_nth_line_from_file(abs_data_file_name, int(exp_dict['header_row']), int(exp_dict['last_row']))
    if names is not None:
        names = names.split(exp_dict['separator'])
    else:
        names = [i for i in range(columnNumber)]

    max_col = min(len(names), len(exp_dict['mapping']))
    obj_map.setNumCols(max_col)
    experiment.setNumColumns(max_col)
    for i in range(max_col):
        mapping = _get_first_column(exp_dict['mapping'], names[i])
        if mapping is None:
            obj_map.setRole(i, COPASI.CExperiment.ignore)
            continue
        role = _role_to_int(mapping['type'])
        obj_map.setRole(i, role)
        need_initial = True if mapping['type'] == 'independent' else False
        cn = mapping['cn'] if 'cn' in mapping else \
                basico.model_info.get_cn(mapping['object'], initial=need_initial ) if 'object' in mapping else \
                None
        if cn is not None:
            obj_map.setObjectCN(i, COPASI.CRegisteredCommonName(cn))
        if role == COPASI.CExperiment.dependent and 'weight' in mapping:
            obj_map.setScale(i, float(mapping['weight']))

    exp_set.addExperiment(experiment)
    #experiment.compile(model.getModel().getMathContainer())
    #experiment.calculateWeights()
    #experiment.updateFittedPoints()





from math import exp
import unittest
import basico

class TestLyap(unittest.TestCase):
    def test_lyap(self):
        dm = basico.load_example('brusselator')
        self.assertIsNotNone(dm)
        exponents, sums, divergence = basico.run_lyapunov(num_exponents=2)
        self.assertIsNotNone(len(exponents), 2)
        self.assertAlmostEqual(sums, (exponents[0] + exponents[1]))
        basico.remove_datamodel(dm)

if __name__ == '__main__':
    unittest.main()



import logging
import subprocess
from multiprocessing import Pool, Lock
from functools import partial
import os

COPASI_SE = 'CopasiSE'
logger = logging.getLogger(__name__)


def process_files(files, pool_size=4, copasi_se=COPASI_SE, max_time=None):
    """Executes all the given files in a multiprocessing pool

    :param files: list of copasi filenames
    :param pool_size: pool size for the multiprocessing pool that will be created (defaults to 4)
    :param copasi_se: path to the copasi SE executable to execute
    :param max_time: maximum time to allow for the execution of each file in seconds (defaults to None for no maximum

    """
    logger.debug(f'Processing {len(files)} files with a pool of size {pool_size}')

    # for debugging purposes use the following line instead of the multiprocessing pool
    # from multiprocessing.dummy import Pool

    with Pool(pool_size) as p:
        p.map(partial(process_file_with_se, copasi_se=copasi_se, max_time=max_time), files)
    logger.debug(f'Processing complete')


def process_file_with_se(filename, copasi_se=COPASI_SE, max_time=None):
    """Executes the specified copasi file with CopasiSE

    :param filename: copasi file containing executable tasks to run
    :param copasi_se: path to the copasi SE executable to execute
    :param max_time: maximum time to allow for the execution of each file in seconds (defaults to None for no maximum)

    """

    args = [copasi_se, '--nologo']
    if max_time is not None:
        args += ['--maxTime', str(max_time)]
    args.append(filename)
    logger.info(f"processing: {filename} with args: {args}")
    subprocess.call(args, cwd=os.path.dirname(filename))



"""Utility methods for working with the mca task

"""
import logging

import basico
import COPASI

from . import model_io

logger = logging.getLogger(__name__)

def run_mca(**kwargs):
    """Runs the mca task, the result is obtained

    by calling:

    * `get_elasticities`
    * `get_flux_control_coefficients` or
    * `get_concentration_control_coefficients`

    :param kwargs: optional arguments

         - | `model`: to specify the data model to be used (if not specified
           | the one from :func:`.get_current_model` will be taken)

         - | `settings`: a dictionary with the settings to apply first

         - | `use_initial_values`: boolean flag indicating whether initial values should be used (true by default)

    :return: None
    """
    model = model_io.get_model_from_dict_or_default(kwargs)
    assert (isinstance(model, COPASI.CDataModel))

    model.getModel().compileIfNecessary()

    if 'settings' in kwargs:
        basico.set_task_settings(basico.T.MCA, kwargs['settings'], model=model)

    task = model.getTask(basico.T.MCA)
    assert (isinstance(task, COPASI.CMCATask))

    if not task.initializeRaw(COPASI.CCopasiTask.OUTPUT_UI):
        logger.error('Could not initialize Sensitivity Task: {0}'.format(COPASI.CCopasiMessage.getAllMessageText()))
        return

    use_initial_values = kwargs.get('use_initial_values', True)

    if not task.processRaw(use_initial_values):
        logger.error('Could not run Sensitivity Task: {0}'.format(COPASI.CCopasiMessage.getAllMessageText()))
        return

    task.restore()



def get_mca_matrix(matrix, scaled=True, run_first=False, **kwargs):
    """Returns the specified mca matrix as pandas data frame

    :param matrix: the matrix to be returned, as string, one of `elasticities`,
                   `flux_control_coefficients` or `concentration_control_coefficients`
    :param scaled: boolean flag indicating whether the scaled matrix should be returned
                   defaults to True

    :param run_first: boolean flag indicating that the task should be run first (defaults to false)

    :param kwargs: optional arguments

         - | `model`: to specify the data model to be used (if not specified
           | the one from :func:`.get_current_model` will be taken)

         - | `settings`: a dictionary with the settings to apply first

         - | `use_initial_values`: boolean flag indicating whether initial values should be used (true by default)

    :return: the selected mca matrix
    :rtype: pd.DataFrame
    """

    model = model_io.get_model_from_dict_or_default(kwargs)
    assert (isinstance(model, COPASI.CDataModel))

    if run_first:
        run_mca(**kwargs)

    task = model.getTask(basico.T.MCA)
    assert (isinstance(task, COPASI.CMCATask))
    method = task.getMethod()
    assert (isinstance(method, COPASI.CMCAMethod))

    if matrix == 'elasticities':
        return basico.model_info._annotated_matrix_to_df(
            method.getScaledElasticitiesAnn() if scaled else method.getUnscaledElasticitiesAnn()
        )

    if matrix == 'concentration_control_coefficients':
        return basico.model_info._annotated_matrix_to_df(
            method.getScaledConcentrationCCAnn() if scaled else method.getUnscaledConcentrationCCAnn()
        )

    if matrix == 'flux_control_coefficients':
        return basico.model_info._annotated_matrix_to_df(
            method.getScaledFluxCCAnn() if scaled else method.getUnscaledFluxCCAnn()
        )

    return None



def get_elasticities(scaled=True, run_first=False, **kwargs):
    """Returns the elasticity matrix as pandas data frame

    :param scaled: boolean flag indicating whether the scaled matrix should be returned
                   defaults to True

    :param run_first: boolean flag indicating that the task should be run first (defaults to false)

    :param kwargs: optional arguments

         - | `model`: to specify the data model to be used (if not specified
           | the one from :func:`.get_current_model` will be taken)

         - | `settings`: a dictionary with the settings to apply first

         - | `use_initial_values`: boolean flag indicating whether initial values should be used (true by default)

    :return: the elasticity matrix
    :rtype: pd.DataFrame
    """
    return get_mca_matrix('elasticities', scaled, run_first, **kwargs)

def get_flux_control_coefficients(scaled=True, run_first=False, **kwargs):
    """Returns the flux control coefficient matrix as pandas data frame

    :param scaled: boolean flag indicating whether the scaled matrix should be returned
                   defaults to True

    :param run_first: boolean flag indicating that the task should be run first (defaults to false)

    :param kwargs: optional arguments

         - | `model`: to specify the data model to be used (if not specified
           | the one from :func:`.get_current_model` will be taken)

         - | `settings`: a dictionary with the settings to apply first

         - | `use_initial_values`: boolean flag indicating whether initial values should be used (true by default)

    :return: the flux control coefficient matrix
    :rtype: pd.DataFrame
    """
    return get_mca_matrix('flux_control_coefficients', scaled, run_first, **kwargs)

def get_concentration_control_coefficients(scaled=True, run_first=False, **kwargs):
    """Returns the concentration control coefficient matrix as pandas data frame

    :param scaled: boolean flag indicating whether the scaled matrix should be returned
                   defaults to True

    :param run_first: boolean flag indicating that the task should be run first (defaults to false)

    :param kwargs: optional arguments

         - | `model`: to specify the data model to be used (if not specified
           | the one from :func:`.get_current_model` will be taken)

         - | `settings`: a dictionary with the settings to apply first

         - | `use_initial_values`: boolean flag indicating whether initial values should be used (true by default)

    :return: the concentration control coefficient matrix
    :rtype: pd.DataFrame
    """
    return get_mca_matrix('concentration_control_coefficients', scaled, run_first, **kwargs)